__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Dict

# import numpy as np
import pandas as pd
import markov_clustering as mc
from umiche.network.CC import CC as gbfscc
from umiche.deduplicate.method.ReformKit import ReformKit as refkit
from umiche.util.Hamming import Hamming
from umiche.network.Adjacency import Adjacency as netadj
from umiche.util.Console import Console


class MarkovClustering:

    def __init__(
            self,
            inflat_val,
            exp_val,
            iter_num,
            verbose=True,
            **kwargs,
    ):
        self.inflat_val = inflat_val
        self.exp_val = exp_val
        self.iter_num = iter_num
        self.kwargs = kwargs
        if 'heterogeneity' in self.kwargs.keys():
            self.heterogeneity = self.kwargs['heterogeneity']
        else:
            self.heterogeneity = False

        self.netadj = netadj()
        self.gbfscc = gbfscc()
        self.refkit = refkit()

        self.console = Console()
        self.console.verbose = verbose

    def adj_to_edge_list(self, graph):
        self.netadj.graph = graph
        return self.netadj.to_edge_list()

    def dfclusters(
            self,
            connected_components,
            graph_adj,
    ):
        """

        Parameters
        ----------
        connected_components
            connected components in dict format:
            {
                'cc0': [...] # nodes,
                'cc1': [...],
                'cc2': [...],
                ...
                'ccn': [...],
            }
            e.g.
            {
                0: ['A', 'B', 'C', 'D', 'E', 'F'],
            }
        graph_adj
            the adjacency list of a graph

        Returns
        -------
            a pandas dataframe
            each connected component is decomposed into more connected subcomponents.

        """
        # print(connected_components)
        # print([*connected_components.values()])
        df_ccs = pd.DataFrame({'cc_vertices': [*connected_components.values()]})
        # print(df_ccs['cc_vertices'])
        df_ccs['graph_cc_adj'] = df_ccs['cc_vertices'].apply(lambda x: self.refkit.graph_cc_adj(x, graph_adj))
        ### @@ graph_cc_adj
        # {'A': ['B', 'C', 'D'], 'B': ['A', 'C'], 'C': ['A', 'B'], 'D': ['A', 'E', 'F'], 'E': ['D'], 'F': ['D']}
        df_ccs['nt_to_int_map'] = df_ccs['graph_cc_adj'].apply(lambda x: self.refkit.keymap(graph_adj=x, reverse=False))
        df_ccs['int_to_nt_map'] = df_ccs['graph_cc_adj'].apply(lambda x: self.refkit.keymap(graph_adj=x, reverse=True))
        ### @@ nt_to_int_map
        # {'A': 0, 'B': 1, 'C': 2, 'D': 3, 'E': 4, 'F': 5}
        ### @@ int_to_nt_map
        # {0: 'A', 1: 'B', 2: 'C', 3: 'D', 4: 'E', 5: 'F'}
        df_ccs['cc_adj_mat'] = df_ccs.apply(
            # lambda x: self.matrix(
            #     graph_adj=x['graph_cc_adj'],
            #     key_map=x['nt_to_int_map'],
            # ),
            lambda x: netadj(graph=x['graph_cc_adj']).to_matrix(),
            axis=1,
        )
        ### @@ cc_adj_mat
        # [[0. 1. 1. 1. 0. 0.]
        #  [1. 0. 1. 0. 0. 0.]
        #  [1. 1. 0. 0. 0. 0.]
        #  [1. 0. 0. 0. 1. 1.]
        #  [0. 0. 0. 1. 0. 0.]
        #  [0. 0. 0. 1. 0. 0.]]
        df_ccs['mcl_clusters'] = df_ccs['cc_adj_mat'].apply(lambda x: self.cluster(x))
        df_ccs['clusters'] = df_ccs.apply(lambda x: self.refkit.key2node(list_2d=x['mcl_clusters'], keymap=x['int_to_nt_map']), axis=1)
        # print(df_ccs['clusters'].values.tolist())
        ### @@ mcl_clusters
        # [(0, 1, 2), (3, 4, 5)]
        ### @@ clusters
        # [['A', 'B', 'C'], ['D', 'E', 'F']]
        df_ccs['clust_num'] = df_ccs['clusters'].apply(lambda x: len(x))
        # df_ccs['clust_num'] = df_ccs['clusters'].apply(lambda x: print(x) if len(x) > 1 else 1)
        ### @@ clust_num
        # 2
        df_ccs['edge_list'] = df_ccs['graph_cc_adj'].apply(lambda graph: self.adj_to_edge_list(graph=graph))
        # print(df_ccs['edge_list'])
        df_ccs['apv'] = df_ccs['edge_list'].apply(lambda edge_list: [list(el) for el in edge_list])
        # print(df_ccs['apv'])
        # df_ccs.apply(lambda x: self.refkit.breakpoint(x, connected_components) if x['clust_num'] > 1 else 1, axis=1)
        return df_ccs

    def dfclusters_cc_all_node_umis(
            self,
            graph_adj,
            int_to_umi_dict,
    ):
        """

        Parameters
        ----------
        connected_components
            connected components in dict format:
            {
                'cc0': [...] # nodes,
                'cc1': [...],
                'cc2': [...],
                ...
                'ccn': [...],
            }
            e.g.
            {
                0: ['A', 'B', 'C', 'D', 'E', 'F'],
            }
        graph_adj
            the adjacency list of a graph

        Returns
        -------
            a pandas dataframe
            each connected component is decomposed into more connected subcomponents.

        """
        df_ccs = pd.DataFrame()
        df_ccs.loc[0, 'method'] = 'dfclusters_cc_all_node_umis'
        onehot_2d_arrs = [list(self.refkit.onehot(umi=umi)) for k, umi in int_to_umi_dict.items()]
        d1 = {'dfclusters_cc_all_node_umis': onehot_2d_arrs}
        d2 = {'dfclusters_cc_all_node_umis': [*int_to_umi_dict.keys()]}
        df_ccs['onehot'] = df_ccs['method'].apply(lambda x: d1[x])
        df_ccs['cc_vertices'] = df_ccs['method'].apply(lambda x: d2[x])
        # print(df_ccs)
        df_ccs['graph_cc_adj'] = df_ccs['cc_vertices'].apply(lambda x: self.refkit.graph_cc_adj(x, graph_adj))
        ### @@ graph_cc_adj
        # {'A': ['B', 'C', 'D'], 'B': ['A', 'C'], 'C': ['A', 'B'], 'D': ['A', 'E', 'F'], 'E': ['D'], 'F': ['D']}
        df_ccs['nt_to_int_map'] = df_ccs['graph_cc_adj'].apply(lambda x: self.refkit.keymap(graph_adj=x, reverse=False))
        df_ccs['int_to_nt_map'] = df_ccs['graph_cc_adj'].apply(lambda x: self.refkit.keymap(graph_adj=x, reverse=True))
        ### @@ nt_to_int_map
        # {'A': 0, 'B': 1, 'C': 2, 'D': 3, 'E': 4, 'F': 5}
        ### @@ int_to_nt_map
        # {0: 'A', 1: 'B', 2: 'C', 3: 'D', 4: 'E', 5: 'F'}
        df_ccs['cc_adj_mat'] = df_ccs.apply(
            # lambda x: self.matrix(
            #     graph_adj=x['graph_cc_adj'],
            #     key_map=x['nt_to_int_map'],
            # ),
            lambda x: netadj(graph=x['graph_cc_adj']).to_matrix(),
            axis=1,
        )
        ### @@ cc_adj_mat
        # [[0. 1. 1. 1. 0. 0.]
        #  [1. 0. 1. 0. 0. 0.]
        #  [1. 1. 0. 0. 0. 0.]
        #  [1. 0. 0. 0. 1. 1.]
        #  [0. 0. 0. 1. 0. 0.]
        #  [0. 0. 0. 1. 0. 0.]]
        df_ccs['mcl_clusters'] = df_ccs['cc_adj_mat'].apply(lambda x: self.cluster(x))
        df_ccs['clusters'] = df_ccs.apply(
            lambda x: self.refkit.key2node(list_2d=x['mcl_clusters'], keymap=x['int_to_nt_map']), axis=1)
        ### @@ mcl_clusters
        # [(0, 1, 2), (3, 4, 5)]
        ### @@ clusters
        # [['A', 'B', 'C'], ['D', 'E', 'F']]
        df_ccs['clust_num'] = df_ccs['clusters'].apply(lambda x: len(x))
        ### @@ clust_num
        # 2

        df_ccs['edge_list'] = df_ccs['graph_cc_adj'].apply(lambda graph: self.adj_to_edge_list(graph=graph))
        # print(df_ccs['edge_list'])
        df_ccs['apv'] = df_ccs['edge_list'].apply(lambda edge_list: [list(el) for el in edge_list])
        # print(df_ccs['apv'])
        return df_ccs

    def cluster(
            self,
            cc_adj_mat,
    ):
        """

        Parameters
        ----------
        cc_adj_mat
            [[0. 1. 1. 1. 0. 0.]
             [1. 0. 1. 0. 0. 0.]
             [1. 1. 0. 0. 0. 0.]
             [1. 0. 0. 0. 1. 1.]
             [0. 0. 0. 1. 0. 0.]
             [0. 0. 0. 1. 0. 0.]]

             # for {'A': ['B', 'C', 'D'], 'B': ['A', 'C'], 'C': ['A', 'B'], 'D': ['A', 'E', 'F'], 'E': ['D'], 'F': ['D']}


        Returns
        -------

        """
        result = mc.run_mcl(
            cc_adj_mat,
            inflation=self.inflat_val,
            expansion=self.exp_val,
            iterations=int(self.iter_num),
        )
        clusters = mc.get_clusters(result)
        # print(clusters)
        return clusters

    def maxval_val(
            self,
            df_mcl_ccs,
            df_umi_uniq_val_cnt,
            thres_fold,
    ):
        """

        Parameters
        ----------
        df_mcl_ccs
        df_umi_uniq_val_cnt
        thres_fold

        Returns
        -------

        """
        # print(df_umi_uniq_val_cnt)
        df_mcl_ccs['mscmv_val'] = df_mcl_ccs['clusters'].apply(
            lambda x: self.maxval_val_(
                mcl_clusters_per_cc=x,
                df_umi_uniq_val_cnt=df_umi_uniq_val_cnt,
                thres_fold=thres_fold,
            )
        )
        df_mcl_ccs['mscmv_val_len'] = df_mcl_ccs['mscmv_val'].apply(lambda x: x[0])
        df_mcl_ccs['mscmv_val_clusters'] = df_mcl_ccs['mscmv_val'].apply(lambda x: x[1])
        df_mcl_ccs['mscmv_val_apv'] = df_mcl_ccs['mscmv_val'].apply(lambda x: x[2])
        df_mcl_ccs['mscmv_val_disapv'] = df_mcl_ccs['mscmv_val'].apply(lambda x: x[3])
        # print(df_mcl_ccs['mscmv_val_len'].sum())
        # print(df_mcl_ccs[['mscmv_val_clusters', 'mscmv_val_disapv', ]])
        # print(df_mcl_ccs['mscmv_val_disapv'],)
        # return {
        #     'count': df_mcl_ccs['mscmv_val_len'],
        #     'clusters': df_mcl_ccs['mscmv_val_clusters'],
        #     'apv': df_mcl_ccs['mscmv_val_apv'],
        #     'disapv': df_mcl_ccs['mscmv_val_disapv'],
        # }
        if self.heterogeneity:
            return (
                df_mcl_ccs['mscmv_val_len'],
                df_mcl_ccs['mscmv_val_clusters'],
                df_mcl_ccs['mscmv_val_apv'],
                df_mcl_ccs['mscmv_val_disapv'],
            )
        else:
            return {
                "count":  df_mcl_ccs['mscmv_val_len'],
                "clusters": df_mcl_ccs['mscmv_val_clusters'],
                "apv": df_mcl_ccs['mscmv_val_apv'],
                "disapv": df_mcl_ccs['mscmv_val_disapv'],
            }

    def maxval_val_(
            self,
            mcl_clusters_per_cc,
            df_umi_uniq_val_cnt,
            thres_fold,
    ):
        """

        Parameters
        ----------
        mcl_clusters_per_cc
        df_umi_uniq_val_cnt
        thres_fold

        Returns
        -------

        """
        mcl_sub_clust_max_val_graph = {}
        mcl_sub_clust_max_val_weights = {}
        for clust in mcl_clusters_per_cc:
            cc_clust_sorted = self.refkit.sort_vals(df_umi_uniq_val_cnt, cc=clust)
            nodes = [*cc_clust_sorted.keys()]
            weights = [*cc_clust_sorted.values()]
            mcl_sub_clust_max_val_graph[nodes[0]] = set()
            mcl_sub_clust_max_val_weights[nodes[0]] = weights[0]
        # print(mcl_sub_clust_max_val_graph)
        approval = []
        disapproval = []
        mscmv_nodes = [*mcl_sub_clust_max_val_weights.keys()]
        mscmv_weights = [*mcl_sub_clust_max_val_weights.values()]
        mscmv_len = len(mscmv_nodes)
        for i in range(mscmv_len):
            for j in range(i + 1, mscmv_len):
                node_i = mscmv_nodes[i]
                node_j = mscmv_nodes[j]
                node_weight_i = mscmv_weights[i]
                node_weight_j = mscmv_weights[j]
                if node_weight_i > thres_fold * node_weight_j - 1:
                    mcl_sub_clust_max_val_graph[node_i].add(node_j)
                    mcl_sub_clust_max_val_graph[node_j].add(node_i)
                    approval.append([node_i, node_j])
                elif node_weight_j > thres_fold * node_weight_i - 1:
                    mcl_sub_clust_max_val_graph[node_i].add(node_j)
                    mcl_sub_clust_max_val_graph[node_j].add(node_i)
                    approval.append([node_i, node_j])
                else:
                    disapproval.append([node_i, node_j])
        # print(mcl_sub_clust_max_val_graph)
        clusters = list(self.gbfscc.deque(mcl_sub_clust_max_val_graph))
        return len(clusters), clusters, approval, disapproval

    def maxval_ed(
            self,
            df_mcl_ccs,
            df_umi_uniq_val_cnt,
            int_to_umi_dict,
            thres_fold,
    ):
        """

        Parameters
        ----------
        df_mcl_ccs
        df_umi_uniq_val_cnt
        int_to_umi_dict
        thres_fold

        Returns
        -------

        """
        df_mcl_ccs['mscmv_ed'] = df_mcl_ccs['clusters'].apply(
            lambda x: self.maxval_ed_(
                mcl_clusters_per_cc=x,
                df_umi_uniq_val_cnt=df_umi_uniq_val_cnt,
                int_to_umi_dict=int_to_umi_dict,
                thres_fold=thres_fold,
            )
        )
        df_mcl_ccs['mscmv_ed_len'] = df_mcl_ccs['mscmv_ed'].apply(lambda x: x[0])
        df_mcl_ccs['mscmv_ed_clusters'] = df_mcl_ccs['mscmv_ed'].apply(lambda x: x[1])
        df_mcl_ccs['mscmv_ed_apv'] = df_mcl_ccs['mscmv_ed'].apply(lambda x: x[2])
        df_mcl_ccs['mscmv_ed_disapv'] = df_mcl_ccs['mscmv_ed'].apply(lambda x: x[3])
        # print(df_mcl_ccs['mscmv_ed_len'].sum())
        # print(df_mcl_ccs[['mscmv_ed_clusters', 'mscmv_ed_disapv', ]])
        # print(df_mcl_ccs['mscmv_ed_clusters'])
        # print('mscmv_ed_apv', df_mcl_ccs['mscmv_ed_apv'])
        # print('mscmv_ed_disapv', df_mcl_ccs['mscmv_ed_disapv'])
        if self.heterogeneity:
            return (
                df_mcl_ccs['mscmv_ed_len'],
                df_mcl_ccs['mscmv_ed_clusters'],
                df_mcl_ccs['mscmv_ed_apv'],
                df_mcl_ccs['mscmv_ed_disapv']
            )
        else:
            return {
                'count': df_mcl_ccs['mscmv_ed_len'],
                'clusters': df_mcl_ccs['mscmv_ed_clusters'],
                'apv': df_mcl_ccs['mscmv_ed_apv'],
                'disapv': df_mcl_ccs['mscmv_ed_disapv'],
            }

    def maxval_ed_(
            self,
            mcl_clusters_per_cc,
            df_umi_uniq_val_cnt,
            int_to_umi_dict,
            thres_fold,
    ):
        """
        # for k1, v1 in mcl_sub_clust_max_val_weights.items():
        #     for k2, v2 in mcl_sub_clust_max_val_weights.items():
        #         if k1 != k2:
        #             edh = Hamming().general(
        #                 int_to_umi_dict[k1],
        #                 int_to_umi_dict[k2],
        #             )
        #             if edh <= thres_fold:
        #                 mcl_sub_clust_max_val_graph[k1].add(k2)
        #                 mcl_sub_clust_max_val_graph[k2].add(k1)
        #                 approval.append([k1, k2])
        #             else:
        #                 disapproval.append([k1, k2])

        Parameters
        ----------
        mcl_clusters_per_cc
            [['A', 'B', 'C'], ['D', 'E', 'F', 'G']]
        df_umi_uniq_val_cnt
            A    120
            D     90
            E     50
            G      5
            B      2
            C      2
            F      1
            dtype: int64
        int_to_umi_dict
            {'A': 'AGATCTCGCA', 'B': 'AGATCCCGCA', 'C': 'AGATCACGCA', 'D': 'AGATCGCGCA', 'E': 'AGATCGCGGA', 'F': 'AGATCGCGTA', 'G': 'TGATCGCGAA'}
        thres_fold
            1

        Returns
        -------

        """
        ### @@ mcl_clusters_per_cc
        # [['A', 'B', 'C'], ['D', 'E', 'F', 'G']]
        ### @@ df_umi_uniq_val_cnt
        # A    120
        # D     90
        # E     50
        # G      5
        # B      2
        # C      2
        # F      1
        # dtype: int64
        ### @@ int_to_umi_dict
        # {'A': 'AGATCTCGCA', 'B': 'AGATCCCGCA', 'C': 'AGATCACGCA', 'D': 'AGATCGCGCA', 'E': 'AGATCGCGGA', 'F': 'AGATCGCGTA', 'G': 'TGATCGCGAA'}
        ### @@ thres_fold
        # 1
        mcl_sub_clust_max_val_graph = {}
        mcl_sub_clust_max_val_weights = {}
        for clust in mcl_clusters_per_cc:
            cc_clust_sorted = self.refkit.sort_vals(df_umi_uniq_val_cnt, cc=clust)
            nodes = [*cc_clust_sorted.keys()]
            weights = [*cc_clust_sorted.values()]
            mcl_sub_clust_max_val_graph[nodes[0]] = set()
            mcl_sub_clust_max_val_weights[nodes[0]] = weights[0]
        # print(mcl_sub_clust_max_val_graph)
        approval = []
        disapproval = []
        mscmv_nodes = [*mcl_sub_clust_max_val_graph.keys()]
        ### @@ mscmv_nodes
        # ['A', 'D']
        mscmv_len = len(mscmv_nodes)
        for i in range(mscmv_len):
            for j in range(i + 1, mscmv_len):
                node_i = mscmv_nodes[i]
                node_j = mscmv_nodes[j]
                edh = Hamming().general(
                    int_to_umi_dict[node_i],
                    int_to_umi_dict[node_j],
                )
                if edh <= thres_fold:
                    mcl_sub_clust_max_val_graph[node_i].add(node_j)
                    mcl_sub_clust_max_val_graph[node_j].add(node_i)
                    approval.append([node_i, node_j])
                else:
                    disapproval.append([node_i, node_j])
        ### @@ mcl_sub_clust_max_val_graph
        # {'A': {'D'}, 'D': {'A'}}
        clusters = list(self.gbfscc.deque(mcl_sub_clust_max_val_graph))
        ### @@ clusters
        # [['A', 'D']]
        return len(clusters), clusters, approval, disapproval

    def decompose(
            self,
            list_nd,
    ) -> Dict:
        """

        Parameters
        ----------
        df

        Returns
        -------
        {

        }

        """
        ### @@ list_nd
        # [list([['A', 'B', 'C'], ['D', 'E', 'F', 'G']])]
        list_md = []
        for i in list_nd:
            ### @@ i
            # [['A', 'B', 'C'], ['D', 'E', 'F', 'G']]
            list_md = list_md + i
        ### @@ list_md
        # [['A', 'B', 'C'], ['D', 'E', 'F', 'G']]
        res = {}
        for i, cc_sub_each_mcl in enumerate(list_md):
            res[i] = cc_sub_each_mcl
        ### @@ res
        # {0: ['A', 'B', 'C'], 1: ['D', 'E', 'F', 'G']}
        return res

    def get_full_subcc(
            self,
            ccs_dict : Dict,
            mcl_ccs_dict : Dict,
    ) -> Dict:
        d = {}
        for ccid2, cc in ccs_dict.items():
            d[ccid2] = []
            for i in cc:
                for ccid1, mcl_cc in mcl_ccs_dict.items():
                    if i in mcl_cc:
                        d[ccid2] = d[ccid2] + mcl_cc
        return d


if __name__ == "__main__":
    from umiche.deduplicate.method.Cluster import Cluster as umimonoclust

    p = MarkovClustering(
        inflat_val=1.6,
        exp_val=2,
        iter_num=100,
    )

    ### @@ data from UMI-tools
    # graph_adj = {
    #     'A': ['B', 'C', 'D'],
    #     'B': ['A', 'C'],
    #     'C': ['A', 'B'],
    #     'D': ['A', 'E', 'F'],
    #     'E': ['D'],
    #     'F': ['D'],
    # }
    # print("An adjacency list of a graph:\n{}".format(graph_adj))
    #
    # node_val_sorted = pd.Series({
    #     'A': 456,
    #     'E': 90,
    #     'D': 72,
    #     'B': 2,
    #     'C': 2,
    #     'F': 1,
    # })
    # print("Counts sorted:\n{}".format(node_val_sorted))

    ### @@ data from mine
    graph_adj = {
        'A': ['B', 'C', 'D'],
        'B': ['A', 'C'],
        'C': ['A', 'B'],
        'D': ['A', 'E', 'F'],
        'E': ['D', 'G'],
        'F': ['D', 'G'],
        'G': ['E', 'F'],
    }
    print("An adjacency list of a graph:\n{}".format(graph_adj))

    node_val_sorted = pd.Series({
        'A': 120,
        'D': 90,
        'E': 50,
        'G': 5,
        'B': 2,
        'C': 2,
        'F': 1,
    })
    print("Counts sorted:\n{}".format(node_val_sorted))

    int_to_umi_dict = {
        'A': 'AGATCTCGCA',
        'B': 'AGATCCCGCA',
        'C': 'AGATCACGCA',
        'D': 'AGATCGCGCA',
        'E': 'AGATCGCGGA',
        'F': 'AGATCGCGTA',
        'G': 'TGATCGCGAA',
    }

    ccs = umimonoclust().cc(graph_adj=graph_adj)
    print("Connected components:\n{}".format(ccs))

    df = p.dfclusters(
        connected_components=ccs,
        graph_adj=graph_adj,
    )
    print(df.columns)
    print("vertices: {}".format(df.loc[0, 'cc_vertices']))
    print(df.loc[0, 'graph_cc_adj'])
    print(df.loc[0, 'nt_to_int_map'])
    print(df.loc[0, 'int_to_nt_map'])
    print(df.loc[0, 'cc_adj_mat'])
    print(df.loc[0, 'mcl_clusters'])
    print(df.loc[0, 'clusters'])
    print(df.loc[0, 'clust_num'])

    print(df['clusters'].values)
    df_mcl_decomposed = p.decompose(list_nd=df['clusters'].values)
    print("deduplicated clusters decomposed (mcl):\n{}".format(df_mcl_decomposed))

    ### @@@ mcl_val
    # df_mcl_val = p.maxval_val(
    #     df_mcl_ccs=df,
    #     df_umi_uniq_val_cnt=node_val_sorted,
    #     thres_fold=2,
    # )
    # print(df_mcl_val)
    # dedup_count = df_mcl_val['count']
    # dedup_clusters = df_mcl_val['clusters']
    # print("deduplicated count (mcl_val):\n{}".format(dedup_count))
    # print("deduplicated clusters (mcl_val):\n{}".format(dedup_clusters))
    #
    # df_mcl_val = p.decompose(list_nd=df_mcl_val['clusters'].values)
    # print("deduplicated clusters decomposed (mcl_val):\n{}".format(df_mcl_val))

    ### @@ mcl_ed
    # df_mcl_ed = p.maxval_ed(
    #     df_mcl_ccs=df,
    #     df_umi_uniq_val_cnt=node_val_sorted,
    #     thres_fold=1,
    #     int_to_umi_dict=int_to_umi_dict,
    # )
    # dedup_count = df_mcl_ed['count']
    # dedup_clusters = df_mcl_ed['clusters']
    # print('approval: {}'.format(df_mcl_ed['apv']))
    #
    # print("deduplicated count (mcl_ed):\n{}".format(dedup_count))
    # print("deduplicated clusters (mcl_ed):\n{}".format(dedup_clusters))
    #
    # df_mcl_ed = p.decompose(list_nd=df_mcl_ed['clusters'].values)
    # print("deduplicated clusters decomposed (mcl_ed):\n{}".format(df_mcl_ed))



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"



class Single:

    def trim(action):
        def tube(deal):
            def wrapper(self, **kwargs):
                if action == 'leave_one_out':
                    res = deal(self, kwargs['ele_loo'])
                    res.remove(kwargs['ele_loo'])
                    return res
                elif action == 'leave_list_out':
                    import numpy as np
                    res = deal(self, kwargs['lis_loo'])
                    ref = np.arange(len(res))
                    c = list(set(ref).difference(set(kwargs['lis_loo'])))
                    # print(c)
                    res_ = np.array(res)
                    res_ = res_[c]
                    return res_
                else:
                    res = deal(self)
                    return res
            return wrapper
        return tube

    def _get(self, gap=False, universal=False):
        """

        Parameters
        ----------
        gap
        universal

        Returns
        -------

        """
        if universal:
            if gap:
                return ['A', 'C', 'G', 'T', '-']
            else:
                return ['A', 'C', 'G', 'T']
        else:
            if gap:
                return ['A', 'T', 'C', 'G', '-']
            else:
                return ['A', 'T', 'C', 'G']

    @trim(action='normal')
    def get(self, gap=False, universal=False):
        """

        Parameters
        ----------
        gap
        universal

        Returns
        -------

        """
        return self._get(gap=gap, universal=universal)

    @trim(action='leave_one_out')
    def getEleTrimmed(self, ele_loo, gap=False, universal=False):
        """

        Parameters
        ----------
        ele_loo
        gap
        universal

        Returns
        -------

        """
        return self._get(gap=gap, universal=universal)

    @trim(action='leave_list_out')
    def getLisTrimmed(self, lis_loo=[], gap=False, universal=False):
        """

        Parameters
        ----------
        lis_loo
        gap
        universal

        Returns
        -------

        """
        return self._get(gap=gap, universal=universal)

    def todict(self, nucleotides, reverse=False):
        """

        Parameters
        ----------
        nucleotides
        reverse

        Returns
        -------

        """
        aa_dict = {}
        for k, v in enumerate(nucleotides):
            aa_dict[v] = k
        if reverse:
            aa_dict = {v: k for k, v in aa_dict.items()}
        return aa_dict

    trim = staticmethod(trim)


if __name__ == "__main__":
    p = Single()
    # bs = p.get()
    bs = p.getEleTrimmed(ele_loo='A')
    # bs = p.getLisTrimmed(lis_loo=[0, 1])
    print(bs)
    print(p.todict(bs, reverse=True))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import pandas as pd
from umiche.fastq.Reader import Reader as rfastq
from umiche.fastq.Writer import Writer as wfastq
from umiche.trim.Filter import Filter
from umiche.trim.BCRuleOut import bcRuleOut as bcro
from umiche.trim.UMIRuleOut import umiRuleOut as umiro
from umiche.trim.SeqRuleOut import seqRuleOut as seqro
from umiche.util.Console import Console


class Template:

    def __init__(
            self,
            params,
            verbose=True,
    ):
        self.params = params
        self.verbose = verbose
        self.filter = Filter()
        self.bcro = bcro(read_summary=self.params, verbose=self.verbose)
        self.umiro = umiro(read_summary=self.params, verbose=self.verbose)
        self.seqro = seqro(read_summary=self.params, verbose=self.verbose)
        self.rfastq = rfastq
        self.wfastq = wfastq

        self.console = Console()
        self.console.verbose = self.verbose

    def todf(self, ):
        """
        Notes
        -----
        Trimmed UMIs and genomic sequences.

        Examples
        --------
        >>> params = {
        ...  'read_struct': 'primer_1+umi_1+seq_1+seq_2+umi_2+primer_2',
        ...  'umi_1': {'len': 12},
        ...  'umi_2': {'len': 10},
        ...  'primer_1': {'len': 20},
        ...  'primer_2': {'len': 20},
        ...  'seq_1': {'len': 6},
        ...  'seq_2': {'len': 8},
        ...  'fastq': {
        ...      'path': to('example/data/'),
        ...      'name': 'pcr_1',
        ...      'trimmed_path': to('example/data/'),
        ...      'trimmed_name': 'pcr_1_trim',
        ...   },
        ... }
        >>> p = template(params)
        >>> p.todf()

        Returns
        -------
            pandas dataframe with column names: seq_raw, name, umi_1, umi_2, ..., umi_n, seq_2, seq_3, ..., seq_n

        """
        self.console.print('===>reading from fastq...')
        names, seqs, _, _ = self.rfastq().fromgz(
            fastq_fpn=self.params['fastq']['fpn'],
        )
        df = pd.DataFrame(seqs, columns=['seq_raw'])
        df['name'] = names
        print('before trimmed', df)
        self.console.print('===>umi structure: {}'.format(self.params['read_struct']))
        compo_struct = self.params['read_struct'].split('+')
        bc_pos_in_struct = [i for i, cstruct in enumerate(compo_struct) if 'bc' in cstruct.split('_')]
        # print(bc_pos_in_struct)
        umi_pos_in_struct = [i for i, cstruct in enumerate(compo_struct) if 'umi' in cstruct.split('_')]
        seq_pos_in_struct = [i for i, cstruct in enumerate(compo_struct) if 'seq' in cstruct.split('_')]
        self.console.print('===>bc positions in the read structure: {}'.format(', '.join(map(str, bc_pos_in_struct))))
        self.console.print('===>umi positions in the read structure: {}'.format(', '.join(map(str, umi_pos_in_struct))))
        self.console.print('===>seq positions in the read structure: {}'.format(', '.join(map(str, seq_pos_in_struct))))
        bc_rule_out_len_dict = self.bcro.sequential(
            compo_struct=compo_struct,
            bc_pos_in_struct=bc_pos_in_struct,
        )
        # print(bc_rule_out_len_dict)
        umi_rule_out_len_dict = self.umiro.sequential(
            compo_struct=compo_struct,
            umi_pos_in_struct=umi_pos_in_struct,
        )
        # print(df)
        seq_rule_out_len_dict = self.seqro.sequential(
            compo_struct=compo_struct,
            seq_pos_in_struct=seq_pos_in_struct,
        )
        # print(df)
        for key, val in bc_rule_out_len_dict.items():
            start = val
            end = val + self.params[key]['len']
            print('BC: start: {} end: {}'.format(start, end))
            df[key] = df.apply(lambda x: self.filter.singleStart(x['seq_raw'], start, end), axis=1)
            self.console.print('===>{} has been taken out'.format(key))
        # print(df)
        for key, val in umi_rule_out_len_dict.items():
            start = val
            end = val + self.params[key]['len']
            print('UMI: start: {} end: {}'.format(start, end))
            df[key] = df.apply(lambda x: self.filter.singleStart(x['seq_raw'], start, end), axis=1)
            self.console.print('===>{} has been taken out'.format(key))
        # print(df)
        for key, val in seq_rule_out_len_dict.items():
            start = val
            end = val + self.params[key]['len']
            df[key] = df.apply(lambda x: self.filter.singleStart(x['seq_raw'], start, end), axis=1)
            self.console.print('===>{} has been taken out'.format(key))
        print('after trimmed', df)
        return df

    def togz(self, df):
        """

        Note
        ----
            write to a Fastq file in gz format.

        Example
        -------
        >>>params = {
        ... 'read_struct': 'primer_1+umi_1+seq_1+seq_2+umi_2+primer_2',
        ... 'umi_1': {'len': 12},
        ... 'umi_2': {'len': 10},
        ... 'primer_1': {'len': 20},
        ... 'primer_2': {'len': 20},
        ... 'seq_1': {'len': 6},
        ... 'seq_2': {'len': 8},
        ... 'fastq': {
        ...     'path': to('example/data/'),
        ...     'name': 'pcr_1',
        ...     'trimmed_path': to('example/data/'),
        ...     'trimmed_name': 'pcr_1_trim',
        ... },
        ...}
        >>>p = Template(params)
        >>>df = p.todf()
        >>>p.togz(df)

        Parameters
        ----------
        df
            pandas Dataframe returned from todf().

        Returns
        -------
            pandas dataframe in the column orders: seq_1, name, umi_1, umi_2, ..., umi_n, seq_2, seq_3, ..., seq_n

        """
        self.console.print('===>start saving in gz format...')
        # # /* block. in case of only UMIs present in the reads. */
        if 'seq_1' not in df.columns.tolist():
            df['seq_1'] = 'B'
        df_col_names = df.columns.tolist()
        self.console.print(df_col_names)
        df_col_names.remove('seq_raw')
        rv_todo_list = [struct for i, struct in enumerate(df_col_names) if 'seq' in struct.split('_')]
        for i in rv_todo_list:
            df_col_names.remove(i)
        if len(rv_todo_list) != 1:
            df['seq'] = ''
            for i in rv_todo_list:
                df['seq'] = df.apply(lambda x: x['seq'] + x[i], axis=1)
        else:
            df['seq'] = df['seq_1']
        df_trimmed = df[['seq'] + df_col_names]
        self.wfastq().togz(
            list_2d=df_trimmed.values.tolist(),
            sv_fpn=self.params['fastq']['trimmed_fpn'],
        )
        self.console.print('===>trimmed UMIs have been saved in gz format.')
        # return df_trimmed


if __name__ == "__main__":
    from umiche.path import to

    params = {
        'umi_1': {
            'len': 12,
        },
        'umi_2': {
            'len': 4,
        },
        'umi_3': {
            'len': 5,
        },
        'bc_1': {
            'len': 2,
        },
        'read_struct': 'umi_1',
        # 'read_struct': 'umi_1+seq_1',
        # 'read_struct': 'bc_1+umi_1+seq_1',
        # 'read_struct': 'primer_1+umi_1+seq_1+seq_2+umi_2+umi_3',
        # 'read_struct': 'primer+umi_start+seq+umi_1+primer',
        # 'read_struct': 'primer+umi_1+seq_1+primer+umi_2+seq_2+umi_3+primer',
        'primer_1': {
            'len': 20,
        },
        'primer_2': {
            'len': 20,
        },
        'seq_1': {
            'len': 6,
        },
        'seq_2': {
            'len': 8,
        },
        'fastq': {
            'fpn': to('data/simu/mclumi/seq_errs/permute_0/seq_err_5.fastq.gz'),
            'trimmed_fpn': to('data/simu/mclumi/seq_errs/permute_0/seq_err_5_trimmed.fastq.gz'),
        },
    }
    p = Template(params)

    df = p.todf()

    print(df)
    print(p.togz(df))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Dict

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from umiche.plot.gadget.Transmitter import Transmitter as transmitter


class TraceMultiple:

    def __init__(
            self,
            df_apv: pd.DataFrame,
            df_disapv: pd.DataFrame,
            scenarios,
            methods,
    ):
        self.df_apv = df_apv
        self.df_disapv = df_disapv
        self.scenarios = scenarios
        self.methods = methods

        sns.set(font="Helvetica")
        sns.set_style("ticks")

    def line(
            self,
            num_img_row=2,
            num_img_col=2,
    ):
        print(self.df_apv)
        print(self.df_disapv)
        fig, ax = plt.subplots(nrows=num_img_row, ncols=num_img_col, figsize=(12, 9), sharey=False, sharex=True)
        palette = sns.color_palette("Paired")

        dict_df = {
            "same origin (merged)": self.df_apv[['same_origin', 'metric', 'method']].rename(columns={"same_origin": 'val'}),
            "different origin (merged)": self.df_apv[['diff_origin', 'metric', 'method']].rename(columns={"diff_origin": 'val'}),
            "same origin (not merged)": self.df_disapv[['same_origin', 'metric', 'method']].rename(columns={"same_origin": 'val'}),
            "different origin (not merged)": self.df_disapv[['diff_origin', 'metric', 'method']].rename(columns={"diff_origin": 'val'}),
        }

        for n, (situation, df) in enumerate(dict_df.items()):

            for j, (method, method_formal) in enumerate(self.methods.items()):
                print(df)
                df_met = df[df['method'] == method_formal]

                if int(n / num_img_col) == num_img_row - 1:
                    xl_mark = True
                elif (int(n / num_img_col) == num_img_row - 2) and (
                        n % num_img_col >= (num_img_col - num_img_col * num_img_row + len([1]))):
                    xl_mark = True
                else:
                    xl_mark = False
                self.line_gadget(
                    ax=ax[int(n / num_img_col), n % num_img_col],
                    x=df_met.index,
                    y=df_met.val.values,
                    line_color=palette[j+1],
                    label=method_formal,  # " ".join(ds_key.split("_"))
                    linewidth=2,
                    marker="o",
                    marker_size=6,
                    marker_edge_width=1.2,
                    marker_face_color='none',
                    decoration_mark=True,
                    xl_mark=xl_mark,
                    yl_mark=True if n % num_img_col == 0 else False,
                    title='{}'.format(situation),
                    title_fs=16,
                    x_label='PCR cycle', # 'PCR cycle'
                    y_label="Percentage of UMI counts",
                    x_label_rotation=0,
                    x_label_rotation_align='right',
                    x_ticks=(np.arange(df_met.index.shape[0])).tolist(),
                    x_ticklabels=df_met['metric'].values.tolist(),
                    x_ticklabel_fs=12,
                    x_label_fs=14,
                    y_label_fs=18,
                    legend_fs=11,
                    legend_loc=None,
                )
        for i in range(num_img_col * num_img_row - (n + 1)):
            ax[num_img_row - 1, num_img_col - 1 - i].set_visible(False)
        plt.subplots_adjust(
            top=0.96,
            bottom=0.06,
            left=0.08,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        plt.show()

    def line_apv(
            self,
            num_img_row=2,
            num_img_col=2,
    ):
        print(self.df_apv)
        fig, ax = plt.subplots(nrows=num_img_row, ncols=num_img_col, figsize=(12, 9), sharey=False, sharex=False)
        palette = sns.color_palette("Paired")

        dict_df = {
            "same origin (merged)": self.df_apv[['same_origin', 'metric', 'method']].rename(columns={"same_origin": 'val'}),
            "different origin (merged)": self.df_apv[['diff_origin', 'metric', 'method']].rename(columns={"diff_origin": 'val'}),
        }

        for n, (situation, df) in enumerate(dict_df.items()):

            for j, (method, method_formal) in enumerate(self.methods.items()):
                print(df)
                df_met = df[df['method'] == method_formal]

                if int(n / num_img_col) == num_img_row - 1:
                    xl_mark = True
                elif (int(n / num_img_col) == num_img_row - 2) and (
                        n % num_img_col >= (num_img_col - num_img_col * num_img_row + len([1]))):
                    xl_mark = True
                else:
                    xl_mark = False
                self.line_gadget(
                    ax=ax[int(n / num_img_col), n % num_img_col],
                    x=df_met.index,
                    y=df_met.val.values,
                    line_color=palette[j+1],
                    label=method_formal,  # " ".join(ds_key.split("_"))
                    linewidth=2,
                    marker="o",
                    marker_size=6,
                    marker_edge_width=1.2,
                    marker_face_color='none',
                    decoration_mark=True,
                    xl_mark=xl_mark,
                    yl_mark=True if n % num_img_col == 0 else False,
                    title='{}'.format(situation),
                    title_fs=16,
                    x_label='PCR cycle', # 'PCR cycle'
                    # y_label="Percentage of UMI counts",
                    y_label="UMI count",
                    x_label_rotation=0,
                    x_label_rotation_align='right',
                    x_ticks=(np.arange(df_met.index.shape[0])).tolist(),
                    x_ticklabels=df_met['metric'].values.tolist(),
                    x_ticklabel_fs=12,
                    x_label_fs=14,
                    y_label_fs=18,
                    legend_fs=11,
                    legend_loc=None,
                )
        for i in range(num_img_col * num_img_row - (n + 1)):
            ax[num_img_row - 1, num_img_col - 1 - i].set_visible(False)
        plt.subplots_adjust(
            top=0.96,
            bottom=0.06,
            left=0.08,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        plt.show()

    @transmitter(type="line", task=None)
    def line_gadget(*args, **kwargs):
        return kwargs

    @transmitter(type="line_scatter", task=None)
    def line_scatter_gadget(*args, **kwargs):
        return kwargs


if __name__ == "__main__":
    from umiche.path import to

    from umiche.deduplicate.io.Stat import Stat as dedupstat

    scenarios = {
        'pcr_nums': 'PCR cycle',
        # 'pcr_errs': 'PCR error',
        # 'seq_errs': 'Sequencing error',
        # 'ampl_rates': 'Amplification rate',
        # 'umi_lens': 'UMI length',
        # 'seq_deps': 'Sequencing depth',
    }
    methods = {
        'unique': 'Unique',
        'cluster': 'Cluster',
        'adjacency': 'Adjacency',
        'directional': 'Directional',
        'dbscan_seq_onehot': 'DBSCAN',
        'birch_seq_onehot': 'Birch',
        'aprop_seq_onehot': 'Affinity Propagation',
        'mcl': 'MCL',
        'mcl_val': 'MCL-val',
        'mcl_ed': 'MCL-ed',
    }

    dedupstat = dedupstat(
        scenarios=scenarios,
        methods=methods,
        param_fpn=to('data/params.yml'),
    )

    p = TraceMultiple(
        df_apv=dedupstat.df_trace_cnt['apv'],
        df_disapv=dedupstat.df_trace_cnt['disapv'],
        scenarios=scenarios,
        methods=methods,
    )

    # p.line()
    p.line_apv()



from .Adjacency import *
from .CC import *
from .Edge import *

# try:
#     from umiche.network.Adjacency import *
#     from umiche.network.CC import *
#     from umiche.network.Edge import *
# except ImportError:
#     pass


# ## /*** block. remote ***/
from .Reader import *
from .Hamming import *
from .Number import *
from .Single import *
from .Writer import *
from .Folder import *
from .Console import *

# ## /*** block. local ***/
# try:
#     from umiche.util.Reader import *
#     from umiche.util.Hamming import *
#     from umiche.util.Number import *
#     from umiche.util.Single import *
#     from umiche.util.Writer import *
#     from umiche.util.Folder import *
# except ImportError:
#     pass


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import numpy as np
import pandas as pd
from umiche.network.Edge import Edge as guuedge
from umiche.util.Hamming import Hamming
from umiche.util.Console import Console


class Build:

    def __init__(
            self,
            df,
            ed_thres,
            verbose=False,
    ):
        """

        Parameters
        ----------
        df
        ed_thres
        """
        self.df = df
        # print(df)
        self.hamming = Hamming()
        self.guuedge = guuedge(verbose=False)
        self.console = Console()
        self.console.verbose = verbose

        self.char_to_int = {'A': 0, 'C': 1, 'G': 2, 'T': 3}

        umi_keymap_stime = time.time()
        self.df_umi_uniq = df.drop_duplicates(subset=['umi'], keep='first')
        ### @@ self.df_umi_uniq
        #           id                     query_name  ...        umi  source
        # 0          0   SRR2057595.2985267_ACCGGTTTA  ...  ACCGGTTTA       1
        # 1          1  SRR2057595.13520751_CCAGGTTCT  ...  CCAGGTTCT       1
        # 2          2   SRR2057595.8901432_AGCGGTTAC  ...  AGCGGTTAC       1
        # ...      ...                            ...  ...        ...     ...
        # 20609  20609   SRR2057595.5197759_TAGGTTCCT  ...  TAGGTTCCT       1
        # [1949 rows x 13 columns]
        self.umi_to_bamid_dict = pd.Series(self.df_umi_uniq.index, index=self.df_umi_uniq.umi.values).to_dict()
        ### @@ self.umi_to_bamid_dict
        # {'ACCGGTTTA': 0, 'CCAGGTTCT': 1, 'AGCGGTTAC': 2, ..., 'TAGGTTCCT': 20609}
        self.uniq_umis = self.df_umi_uniq['umi'].values
        # print(self.uniq_umis)
        ### @@ self.uniq_umis
        # ['ACCGGTTTA' 'CCAGGTTCT' 'AGCGGTTAC', ..., 'TAGGTTCCT']
        # self.uniq_umis1 = self.df['umi'].unique()
        self.uniq_umi_num = self.uniq_umis.shape[0]
        self.console.print('=========># of unique UMIs: {}'.format(self.uniq_umi_num))
        self.umi_to_int_dict = {k: id for id, k in enumerate(self.uniq_umis)}
        ### @@ self.umi_to_int_dict
        # {'ACCGGTTTA': 0, 'CCAGGTTCT': 1, 'AGCGGTTAC': 2, ..., 'TAGGTTCCT': 1948}
        self.int_to_umi_dict = {id: k for k, id in self.umi_to_int_dict.items()}
        ### @@ self.int_to_umi_dict
        # {0: 'ACCGGTTTA', 1: 'CCAGGTTCT', 2: 'AGCGGTTAC', ..., 1948: 'TAGGTTCCT'}
        self.df_umi_uniq_val_cnt = self.df['umi'].value_counts(ascending=False)
        ### @@ self.df_umi_uniq_val_cnt
        # umi
        # AGCGGTTAC    55
        # TACGGTTAT    51
        # AATGGTTTC    47
        #              ..
        # TAGGTTCCT     1
        df_umi_uniq_val_cnt_ids = self.df_umi_uniq_val_cnt.index
        ### @@ df_umi_uniq_val_cnt_ids
        # Index(['AGCGGTTAC', 'TACGGTTAT', 'AATGGTTTC',  ... 'TAGGTTCCT'], dtype='object', name='umi', length=1949)
        self.df_umi_uniq_val_cnt.index = [self.umi_to_int_dict[i] for i in df_umi_uniq_val_cnt_ids]
        ### @@ self.df_umi_uniq_val_cnt
        # 2       55
        # 780     51
        # 416     47
        #         ..
        # 1948     1
        # Name: count, Length: 1949, dtype: int64
        self.console.print('=========>umi keymap time: {:.3f}s'.format(time.time() - umi_keymap_stime))

        self.uniq_umi_id_to_bam_id_dict = {}
        # self.umi_bam_ids1 = {}
        for k, v in self.int_to_umi_dict.items():
            self.uniq_umi_id_to_bam_id_dict[k] = self.umi_to_bamid_dict[v]
            ### @@ self.umi_to_bamid_dict[v]
            # 0
            # 1
            # 2
            # ...
            # 20609
            ### @@ slow scheme
            # self.umi_bam_ids1[k] = df.loc[df['umi'].isin([v])]['id'].values[0]
        ed_list_stime = time.time()
        self.ed_list = self.ed_list_()
        ### @@ self.ed_list
        # [[0, 1, 4],
        # [0, 2, 3],
        # [0, 3, 2],
        # ...
        # [1947, 1948, 4]]
        self.console.print('=========>edit distance list construction time: {:.3f}s'.format(time.time() - ed_list_stime))
        # self.ed_list = self.calc_ed()
        # print(self.ed_list)
        self.df_eds = pd.DataFrame(self.ed_list, columns=['node_1', 'node_2', 'ed'])
        if len(self.df_eds['ed']):
            self.ave_ed = np.ceil(sum(self.df_eds['ed']) / (len(self.df_eds['ed'])))
        else:
            self.ave_ed = -1
            ### @@ self.ave_ed
        # 5.0
        ### @@ self.df_eds
        #          node_1  node_2  ed
        # 0             0       1   4
        # 1             0       2   3
        # 2             0       3   2
        # ...         ...     ...  ..
        # 1898325    1947    1948   4
        # [1898326 rows x 3 columns]
        self.df_ed_sel = self.df_eds.loc[self.df_eds['ed'] == ed_thres]
        # print(self.df_ed_sel['ed'].value_counts())

        ### @@ self.df_ed_sel
        #          node_1  node_2  ed
        # 76            0      77   1
        # 80            0      81   1
        # 96            0      97   1
        # ...         ...     ...  ..
        # 1898293    1940    1944   1
        # [14200 rows x 3 columns]
        self.console.print('=========>edit distance list construction time: {:.3f}s'.format(time.time() - ed_list_stime))

        edge_list_stime = time.time()
        self.edge_list = self.guuedge.fromdf(self.df_ed_sel, to_tuple=False)
        ### @@  self.edge_list
        # [[0, 77], [0, 81], [0, 97], ..., [1940, 1944]]
        self.console.print('=========>edge list construction time: {:.3f}s'.format(time.time() - edge_list_stime))

        graph_adj_stime = time.time()
        self.graph_adj = {i: [] for i in [*self.umi_to_int_dict.values()]}
        ### @@ self.graph_adj
        # {0: [], 1: [], 2: [], ..., 1948: []}
        self.guuedge.graph = self.edge_list
        self.graph_adj_edges = self.guuedge.to_adj_dict()
        ### @@ self.graph_adj_edges
        # {0: [77, 81, 97, 153, 205, 228, 239, 433, 518, 562, 602, 728, 791, 839, 930],
        # 1: [111, 328, 348, 472, 494, 779, 789, 927, 945, 946, 1029, 1055, 1210, 1345, 1398],
        # 2: [156, 163, 168, 215, 282, 422, 480, 565, 765, 831, 843, 1022, 1166, 1229, 1474],
        # ...,
        # 1947: [143, 1066, 1080, 1183, 1199, 1308, 1318, 1606, 1635, 1640, 1682, 1751, 1906, 1935],
        # 1948: [55, 151, 241, 609, 1025, 1032, 1060, 1191, 1230, 1307, 1489, 1490, 1713, 1851, 1922]}
        ### @@ len(self.graph_adj_edges)
        # 1949
        for k, v in self.graph_adj_edges.items():
            self.graph_adj[k] += v
        ### @@ self.graph_adj
        # {0: [77, 81, 97, 153, 205, 228, 239, 433, 518, 562, 602, 728, 791, 839, 930],
        # 1: [111, 328, 348, 472, 494, 779, 789, 927, 945, 946, 1029, 1055, 1210, 1345, 1398],
        # 2: [156, 163, 168, 215, 282, 422, 480, 565, 765, 831, 843, 1022, 1166, 1229, 1474],
        # ...,
        # 1948: [55, 151, 241, 609, 1025, 1032, 1060, 1191, 1230, 1307, 1489, 1490, 1713, 1851, 1922]}

        # from umiche.deduplicate.method.Cluster import cluster as umimonoclust
        # cc = umimonoclust().cc(self.graph_adj)
        # print(cc)
        # print(len(cc))
        # fff = False
        # if len(self.int_to_umi_dict) > 200:
        #     print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
        #     print(self.edge_list)
        #     print(self.graph_adj)
        #     print(self.int_to_umi_dict)
        #     print(self.df_umi_uniq_val_cnt)
        #     fff = True
        # else:
        #     fff = False
            # break
        self.console.print('===>graph adjacency list construction time: {:.3f}s'.format(time.time() - graph_adj_stime))
        self.data_summary = {
            'graph_adj': self.graph_adj,
            'int_to_umi_dict': self.int_to_umi_dict,
            'df_umi_uniq_val_cnt': self.df_umi_uniq_val_cnt,
            'uniq_umi_id_to_bam_id_dict': self.uniq_umi_id_to_bam_id_dict,
            'ave_ed': self.ave_ed if len(self.int_to_umi_dict) != 1 else -1,
            # 'fff': fff,
        }

    def calc_ed(self, ):
        nda_umi_db = np.array([list(umi) for umi in self.uniq_umis])
        nda_umi_db = np.vectorize(self.char_to_int.get)(nda_umi_db)
        df = pd.DataFrame()
        for i in range(self.uniq_umi_num):
            # r_ids = np.arange(i, self.uniq_umi_num)
            # l_map_ids = [self.umi_to_int_dict[self.uniq_umis[i]]] * (self.uniq_umi_num - i)
            # r_map_ids = [self.umi_to_int_dict[self.uniq_umis[r]] for r in r_ids]
            dd = self.np_vectorize(
                # l_map_ids=l_map_ids,
                # r_map_ids=r_map_ids,
                db_ref=nda_umi_db[i:],
                umi=self.uniq_umis[i],
            )
            df = pd.concat([df, dd], axis=0)
        return df

    def np_vectorize(
            self,
            # l_map_ids,
            # r_map_ids,
            db_ref,
            umi,
    ):
        # stime = time.time()
        df = pd.DataFrame()
        si_arr = np.sum(db_ref != [self.char_to_int[s] for s in umi], axis=1)
        # print(si_arr)
        # print("==================>time: {:.5f}s".format(time.time() - stime))
        # df[0] = l_map_ids
        # df[1] = r_map_ids
        df['ed'] = si_arr
        # print(df)
        # print(cnt_dict)
        # with open(self.sv_fp + seq1 + '.json', 'w') as fp:
        #     json.dump(cnt_dict, fp)
        return df

    def ed_list_(self, ):
        eds = []
        for i in range(self.uniq_umi_num):
            for j in range(i + 1, self.uniq_umi_num):
                l = self.uniq_umis[i]
                r = self.uniq_umis[j]
                # if self.umi_to_int_dict[l] == 31:
                #     print(l)
                # if self.umi_to_int_dict[r] == 50:
                #     print(r)
                eds.append([
                    self.umi_to_int_dict[l],
                    self.umi_to_int_dict[r],
                    self.hamming.general(l, r),
                ])
        # print(len(eds))
        return eds


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import pandas as pd
from umiche.fastq.Convert import convert as fas2bam
from umiche.trim.Template import template as umitrim
from umiche.util.Writer import writer as gwriter
from umiche.network.bfs.ConnectedComponent import ConnectedComponent as gbfscc
from umiche.deduplicate.trimer.pipeline import Config
from umiche.deduplicate.MultiPos import MultiPos as deduppos
from umiche.plot.scenario.TraceSingle import valid as plotv
from umiche.path import to


class umi(Config.Config):

    def __init__(
            self,
            metric,
            method,
            comp_cat,
            fastq_fp=None,
            is_trim=False,
            is_tobam=False,
            is_dedup=False,
    ):
        super(umi, self).__init__()
        self.metric = metric
        self.comp_cat = comp_cat
        self.gbfscc = gbfscc()
        self.method = method
        self.gwriter = gwriter()
        self.plotv = plotv()
        df_dedup = pd.DataFrame()
        for i_pn in range(self.permutation_num):
            dedup_arr = []
            for id, i_metric in enumerate(self.metric_vals[self.metric]):
                if self.metric == 'pcr_nums':
                    print('=>at PCR {}'.format(i_metric))
                    fn_surf = str(i_metric)
                    self.mcl_inflat = 2.3
                    self.mcl_exp = 2
                    self.mcl_fold_thres = 1
                    self.umi_len = self.umi_unit_len_fixed * self.umi_unit_pattern
                elif self.metric == 'pcr_errs':
                    self.mcl_inflat = i_metric
                    print('=>No.{} PCR error: {}'.format(id, i_metric))
                    fn_surf = str(id)
                    # # /*** mcl_ed params ***/
                    # self.mcl_inflat = 1.1 if i_metric > 0.005 else 1.7
                    # self.mcl_exp = 2
                    # self.mcl_fold_thres = 1

                    # # /*** mcl_val params ***/
                    self.mcl_inflat = 1.1 if i_metric > 0.005 else 1.8
                    self.mcl_exp = 2
                    self.mcl_fold_thres = 2
                    self.umi_len = self.umi_unit_len_fixed * self.umi_unit_pattern
                elif self.metric == 'seq_errs':
                    print('=>No.{} sequencing error: {}'.format(id, i_metric))
                    # self.mcl_inflat = 1.1 if i_metric > 0.005 else 2.7
                    # self.mcl_exp = 3
                    self.mcl_fold_thres = 1.6
                    self.mcl_inflat = 1.1 if i_metric > 0.005 else 2.7
                    self.mcl_exp = 2
                    fn_surf = str(id)
                    self.umi_len = self.umi_unit_len_fixed * self.umi_unit_pattern
                elif self.metric == 'ampl_rates':
                    print('=>No.{} amplification rate: {}'.format(id, i_metric))
                    fn_surf = str(id)
                    # self.mcl_inflat = 1.3 if i_metric > 0.5 else 2
                    # # /*** mcl_ed params ***/
                    # if i_metric < 8:
                    #     self.mcl_inflat = 4
                    # if i_metric >= 8 and i_metric <= 11:
                    #     self.mcl_inflat = 2.3
                    # if i_metric > 11:
                    #     self.mcl_inflat = 1.1
                    # self.mcl_exp = 3
                    # self.mcl_fold_thres = 1

                    # /*** mcl_val params ***/
                    if i_metric < 8:
                        self.mcl_inflat = 2
                    if i_metric >= 0.9:
                        self.mcl_inflat = 1.8
                    self.mcl_exp = 4
                    self.mcl_fold_thres = 11

                    self.umi_len = self.umi_unit_len_fixed * self.umi_unit_pattern
                elif self.metric == 'umi_lens':
                    print('=>No.{} UMI length: {}'.format(id, i_metric))
                    fn_surf = str(i_metric)
                    # self.mcl_inflat = 1.1 if i_metric > 11 else 2.3
                    # # /*** mcl_ed params ***/
                    # if i_metric < 8:
                    #     self.mcl_inflat = 4
                    # if i_metric >= 8 and i_metric <= 11:
                    #     self.mcl_inflat = 2.3
                    # if i_metric > 11:
                    #     self.mcl_inflat = 1.1
                    # self.mcl_exp = 3
                    # self.mcl_fold_thres = 1

                    # # # /*** mcl_val params ***/
                    # if i_metric < 8:
                    #     self.mcl_inflat = 6
                    # if i_metric >= 8 and i_metric <= 11:
                    #     self.mcl_inflat = 2.3
                    # if i_metric > 11:
                    #     self.mcl_inflat = 1.1
                    # self.mcl_exp = 4
                    # self.mcl_fold_thres = 11

                    # # /*** mcl_val params ***/
                    if i_metric < 8:
                        self.mcl_inflat = 5.8
                        self.mcl_exp = 6
                    if i_metric >= 8 and i_metric <= 11:
                        self.mcl_inflat = 2.3
                        self.mcl_exp = 4
                    if i_metric > 11:
                        self.mcl_inflat = 1.1
                        self.mcl_exp = 4
                    self.mcl_fold_thres = 11

                    self.umi_len = i_metric
                else:
                    fn_surf = str(i_metric)
                    self.umi_len = self.umi_unit_len_fixed * self.umi_unit_pattern
                fn = self.fn_pref[self.metric] + fn_surf
                if is_trim:
                    self.trim(
                        fastq_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/' + fn,
                        fastq_trimmed_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/trimmed/' + fn,
                        umi_len=self.umi_len,
                    )
                if is_tobam:
                    fas2bam(
                        fastq_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/trimmed/' + fn + '.fastq.gz',
                        # fastq_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/' + self.comp_cat + '/' + fn + '.fastq.gz',
                        bam_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/' + self.comp_cat + '/bam/' + fn,
                    ).tobam()
                if is_dedup:
                    # if self.metric == 'seq_errs':
                    #     if i_metric == 0.125 or i_metric == 0.15:
                    #         continue
                    #     else:
                    dedup_ob = deduppos(
                        mode='internal',
                        method=self.method,
                        # bam_fpn=to('example/data/example.bam'),
                        bam_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/' + self.comp_cat + '/bam/' + fn + '.bam',
                        pos_tag='PO',
                        mcl_fold_thres=self.mcl_fold_thres,
                        inflat_val=self.mcl_inflat,
                        exp_val=self.mcl_exp,
                        iter_num=100,
                        verbose=False,
                        ed_thres=1,
                        is_sv=False,
                        sv_fpn=fastq_fp + self.metric + '/permute_' + str(i_pn) + '/' + self.comp_cat + '/' + fn,
                    )
                    dedup_arr.append(dedup_ob.dedup_num)
            df_dedup['pn' + str(i_pn)] = dedup_arr
            print(df_dedup)
        self.gwriter.generic(
            df=df_dedup,
            sv_fpn=fastq_fp + self.metric + '/' + str(self.method) + '_' + self.comp_cat + '.txt',
            header=True,
        )

    def trim(self, fastq_fpn, fastq_trimmed_fpn, umi_len):
        trim_params = {
            'read_struct': 'umi_1',
            'umi_1': {
                'len': umi_len,
            },
            'fastq': {
                'fpn': fastq_fpn + '.fastq.gz',
                'trimmed_fpn': fastq_trimmed_fpn + '.fastq.gz',
            },
        }
        umitrim_parser = umitrim(trim_params)
        df = umitrim_parser.todf()
        umitrim_parser.togz(df)
        return 0


if __name__ == "__main__":
    p = umi(
        # metric='pcr_nums',
        # metric='pcr_errs',
        metric='seq_errs',
        # metric='ampl_rates',
        # metric='umi_lens',

        # method='unique',
        # method='cluster',
        # method='adjacency',
        method='directional',
        # method='mcl',
        # method='mcl_val',
        # method='mcl_ed',
        # method='set_cover',

        comp_cat='',
        # comp_cat='trimmed',
        # comp_cat='lmr',
        # comp_cat='ref',
        # comp_cat='bipartite',
        # comp_cat='double',
        # comp_cat='correct',

        # is_trim=True,
        # is_tobam=False,
        # is_dedup=False,

        # is_trim=False,
        # is_tobam=True,
        # is_dedup=False,

        is_trim=False,
        is_tobam=False,
        is_dedup=True,

        fastq_fp=to('data/simu/tree/trimer/'),
        # fastq_fp=to('data/simu/monomer/pcr8/'),
        # fastq_fp=to('data/simu/trimer/pcr8/'),
        # fastq_fp=to('data/simu/dimer/pcr8/'),
        # fastq_fp=to('data/simu/dimer/treepcr22_250/'),
        # fastq_fp=to('data/simu/dimer/pcr8_mono24/'),
    )



__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import pandas as pd
from umiche.util.Console import Console


class Relation:

    def __init__(
            self,
            df,
            verbose=False,
    ):
        self.console = Console()
        self.console.verbose = verbose

        self.df = df
        self.df['umi#'] = self.df['query_name'].apply(lambda x: x.split('_')[0].split('-')[0])
        # print(self.df)
        ### @@ self.df
        #           id            query_name  flag  ...         umi  source  umi#
        # 0          0   23-pcr-8_GCATCAGTAG     0  ...  GCATCAGTAG       1    23
        # 1          1    6-pcr-8_CTGACTGCGC     0  ...  CTGACTGCGC       1     6
        # 2          2   13-pcr-7_CACACGATAC     0  ...  CACACGATAC       1    13
        # ...      ...                   ...   ...  ...         ...     ...   ...
        # 23813  23813   47-pcr-9_GCGAAGAGGA     0  ...  GCGAAGAGGA       1    47
        self.df['umi_pcr#'] = self.df['query_name'].apply(lambda x: self.pcrnum(x))
        # print(self.df)
        ### @@ self.df
        #           id            query_name  flag  ...  source  umi#  umi_pcr#
        # 0          0   23-pcr-8_GCATCAGTAG     0  ...       1    23         8
        # 1          1    6-pcr-8_CTGACTGCGC     0  ...       1     6         8
        # 2          2   13-pcr-7_CACACGATAC     0  ...       1    13         7
        # ...      ...                   ...   ...  ...     ...   ...       ...
        # 23813  23813   47-pcr-9_GCGAAGAGGA     0  ...       1    47         9
        ### @@@ note version 2 for resimpy
        # self.df['umi_src'] = self.df['query_name'].apply(lambda x: x.split('_')[0].split('-')[1])
        ### @@@ note version 2 for phylotres as well as resimpy
        self.df['umi_src'] = self.df['query_name'].apply(lambda x: x.split('_')[-2].split('-')[1])
        # print(self.df)
        ### @@ self.df
        #           id            query_name  flag  ...  umi#  umi_pcr#  umi_src
        # 0          0   23-pcr-8_GCATCAGTAG     0  ...    23         8      pcr
        # 1          1    6-pcr-8_CTGACTGCGC     0  ...     6         8      pcr
        # 2          2   13-pcr-7_CACACGATAC     0  ...    13         7      pcr
        # ...      ...                   ...   ...  ...   ...       ...      ...
        # 23813  23813   47-pcr-9_GCGAAGAGGA     0  ...    47         9      pcr
        umi_keymap_stime = time.time()
        self.df_umi_uniq = df.drop_duplicates(subset=['umi'], keep='first')
        ### @@ self.df_umi_uniq
        #           id            query_name  flag  ...  umi#  umi_pcr#  umi_src
        # 0          0   23-pcr-8_GCATCAGTAG     0  ...    23         8      pcr
        # 1          1    6-pcr-8_CTGACTGCGC     0  ...     6         8      pcr
        # 2          2   13-pcr-7_CACACGATAC     0  ...    13         7      pcr
        # ...      ...                   ...   ...  ...   ...       ...      ...
        # 23691  23691   23-pcr-8_GCATCAGGAG     0  ...    23         8      pcr
        # [257 rows x 17 columns]
        self.uniq_umis = self.df_umi_uniq['umi'].values
        self.uniq_umi_num = self.uniq_umis.shape[0]
        self.console.print('==================>unique UMI number: {}'.format(self.uniq_umi_num))

        self.umi_to_int_dict = {k: id for id, k in enumerate(self.uniq_umis)}
        ### @@ self.umi_to_int_dict
        # {'GCATCAGTAG': 0, 'CTGACTGCGC': 1, 'CACACGATAC': 2, 'TTAGATGATT': 3, ..., 'GAAGTATATT': 255, 'GCATCAGGAG': 256}
        self.int_to_umi_dict = {id: k for k, id in self.umi_to_int_dict.items()}
        self.df_umi_uniq_val_cnt = self.df['umi'].value_counts(ascending=False)
        ### @@ self.df_umi_uniq_val_cnt
        # umi
        # AGTACGCGAG    636
        # GGAGATCCGG    625
        # AGTAGAACCC    619
        #              ...
        # GCATCAGGAG      1
        # Name: count, Length: 257, dtype: int64
        df_umi_uniq_val_cnt_ids = self.df_umi_uniq_val_cnt.index
        self.df_umi_uniq_val_cnt.index = [self.umi_to_int_dict[i] for i in df_umi_uniq_val_cnt_ids]
        self.console.print('==================>umi keymap time: {:.3f}s'.format(time.time() - umi_keymap_stime))

        umi_trace_dict_stime = time.time()
        self.umi_id_to_origin_id_dict = {}
        umi_str_to_origin_dict = pd.Series(self.df_umi_uniq['umi#'].values, index=self.df_umi_uniq['umi'].values).to_dict()
        ### @@ umi_str_to_origin_dict
        # {'GCATCAGTAG': '23', 'CTGACTGCGC': '6', 'CACACGATAC': '13', 'TTAGATGATT': '10', ..., 'GAAGTATATT': '27', 'GCATCAGGAG': '23'}
        for uniq_umi in self.uniq_umis:
            self.umi_id_to_origin_id_dict[self.umi_to_int_dict[uniq_umi]] = int(umi_str_to_origin_dict[uniq_umi])
        ### @@ self.umi_id_to_origin_id_dict
        # {0: 23, 1: 6, 2: 13, 3: 10, ..., 255: 27, 256: 23}

        self.console.print('==================>umi trace dict time: {:.3f}s'.format(time.time() - umi_trace_dict_stime))

    def pcrnum(self, x):
        # print(x.split('_'))
        ### @@@ note version 1 for resimpy
        # c = x.split('_')[0].split('-')
        ### @@@ note version 2 for phylotres as well as resimpy
        c = x.split('_')[-2].split('-')
        # print(c)
        if c[1] == 'init':
            return -1
        else:
            return c[2]


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np
import tensorflow as tf
from umiche.deepconvcvae.CondiConvVAE import condiConvVAE
from umiche.plot.scatter.DimensionReduction import dimensionReduction as drplot


def sampling(
        decoder,
        num_labels,
        model_name,
        image_size,
        num_dots=1000,
        batch_size=32,
        sv_fpn=None,
):
    xs = []
    ys = []
    for ilabel in range(num_labels):
        print('===>label {}'.format(ilabel))
        digit_size = image_size
        # z_combo_spl = np.random.poisson(5, size=[num_dots, 2])
        z_combo_spl = np.random.normal(loc=1, scale=2, size=[num_dots, 2])
        # z_combo_spl = np.random.normal(loc=1, scale=4, size=[num_dots, 2])
        # z_combo_spl = np.random.uniform(0, 4, size=[num_dots, 2])
        ilabel_ = np.eye(num_labels)[np.array([ilabel])]
        for i, zcspl in enumerate(z_combo_spl):
            z_sample = np.array([zcspl])
            x_decoded = decoder.predict([z_sample, ilabel_])
            # x_decoded = decoder.get_layer("decoder").predict([z_sample, ilabel_])
            # print(x_decoded.shape)
            # print(x_decoded[0].shape)
            digit = x_decoded[0].reshape(digit_size, digit_size)
            # print(digit)
            t = digit.reshape(digit_size * digit_size)*260
            t = np.floor(t).astype(int)
            # print(t[t>1])
            # print(t[t>1].shape)

            xs.append(t)
            # xs.append(x_decoded[0])
            ys.append(ilabel)
            # sds = digit * 255
            # print(np.where(sds > 1, sds, 0))
    print(np.array(xs))
    from scipy.sparse import csr_matrix

    print(csr_matrix(np.array(xs)))
    import pandas as pd
    sdf = pd.DataFrame(np.array(xs))
    sdf['Y'] = ys
    # sdf = pd.DataFrame.sparse.from_spmatrix(csr_matrix(np.array(xs)))
    print(sdf)
    if sv_fpn:
        # sv_h5_fpn = 'data_' + str(num_dots) + '_' + model_name + '.h5'
        sdf.to_hdf(sv_fpn, key='df', mode='w')
        print('saved')
        # print(pd.read_hdf(sv_h5_fpn, 'df'))

    # import pandas as pd
    # df = pd.read_hdf('data.h5', 'df')
    # df = df.drop(columns=['Y'])
    # import numpy as np
    # print(np.unique(df.to_numpy().flatten()))

    # drplot().single(X=xs, y=ys, tech='TSNE', marker_size=3, cmap='tab20b', title='CondiCVAE')
    # drplot().single(X=xs, y=ys, tech='PCA', marker_size=3, cmap='tab20b', title='CondiCVAE')
    # drplot().single(X=xs, y=ys, tech='UMAP', marker_size=3, cmap='tab20b', title='CondiCVAE')


def simulate(
        model_fpn,
        image_size,
        batch_size=32,
        kernel_size=3,
        filters=16,
        latent_dim=2,
        strides=2,
        epochs=1,
        num_dots=100,
        model_name='cvae_model50',
        sv_fpn=None,
):
    # model_name = "./cvae_model20"
    # save_dir = "./cvae_model20"
    # if not os.path.isdir(save_dir):
    #     os.makedirs(save_dir)
    # filename = model_name + '.tf'
    # filepath = os.path.join(save_dir, filename)
    # # ccvae.save_weights(filepath)
    # print(filepath)
    num_labels = 11
    input_shape = (image_size, image_size, 1)
    label_shape = (num_labels, )

    inputs = tf.keras.layers.Input(shape=input_shape, name='encoder_input')
    y_labels = tf.keras.layers.Input(shape=label_shape, name='class_labels')
    print(inputs)
    print(y_labels)
    condicvae = condiConvVAE(
        image_size=image_size,
        input_shape=input_shape,
        label_shape=label_shape,
        batch_size=batch_size,
        kernel_size=kernel_size,
        filters=filters,
        latent_dim=latent_dim,
        strides=strides,
        epochs=epochs,
        inputs=inputs,
        y_labels=y_labels,
    )
    shape, z_mean, z_log_var, encoder = condicvae.encoding()
    # encoder.summary()
    decoder = condicvae.decoding(shape)
    # decoder.summary()
    # plot_model(decoder, to_file='cvae_cnn_decoder.png', show_shapes=True)

    outputs = decoder([encoder([inputs, y_labels])[2], y_labels])
    ccvae = tf.keras.models.Model(inputs=[inputs, y_labels], outputs=outputs, name='ccvae')

    ccvae.load_weights(
        model_fpn,
    )

    sampling(
        decoder,
        num_labels=num_labels,
        model_name=model_name,
        image_size=image_size,
        num_dots=num_dots,
        batch_size=32,
        sv_fpn=sv_fpn,
    )


if __name__ == "__main__":
    print(simulate(
        image_size=144,
        batch_size=32,
        kernel_size=3,
        filters=16,
        latent_dim=2,
        num_dots=100,
        strides=2,
        epochs=1,
        model_fpn='../data/deepconvcvae/cvae_model50/cvae_model50.tf',
        sv_fpn='../data/deepconvcvae/cvae_model50/cvae_model50.h5',
    ))


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import time
import pandas as pd
from umiche.util.Console import Console


class Reader:

    def __init__(
            self,
            bam_fpn,
            verbose=False,
    ):
        """
        # pos_last = 0
        # chr_last = 0
        # if read.pos <= (pos_last + 1000) and read.reference_id == chr_last:
        # print(read.pos, pos_last, read.reference_id)
        # pos_last = read.pos
        # chr_last = read.reference_id
        Parameters
        ----------
        bam_fpn
        verbose
        """
        import pysam
        # import bamnostic as bs

        self.console = Console()
        self.console.verbose = verbose
        self.console.print('===>reading the bam file... {}'.format(bam_fpn))
        read_bam_stime = time.time()
        self.pysam_bam = pysam.AlignmentFile(bam_fpn, "rb")
        # self.pysam_bam = bs.AlignmentFile(bam_fpn, "rb")
        self.console.print('===>reading BAM time: {:.2f}s'.format(time.time() - read_bam_stime))

    def bycol(self, col='sname'):
        """

        Parameters
        ----------
        col

        Returns
        -------

        """
        t = []
        if col == 'sname':
            for id, read in enumerate(self.pysam_bam):
                # print(read)
                # print(read.get_tags())
                if read.reference_id != -1:
                    tt = read.query_name
                else:
                    continue
                t.append(tt)
            return pd.DataFrame(t)

    def todf(
            self,
            tags=[],
            filter_by=None,
    ):
        """

        Notes
        -----
        11 columns from alignments deciphered by Pysam
            # read.query_name
            # read.flag
            # read.reference_id
            # read.reference_start
            # read.mapping_quality
            # read.cigar
            # read.query_sequence
            # read.next_reference_id
            # read.next_reference_start
            # read.template_length
            # read.query_qualities

        See Also
        --------
        https://pysam.readthedocs.io/en/latest/usage.html#creating-bam-cram-sam-files-from-scratch
        https://pysam.readthedocs.io/_/downloads/en/v0.12.0/pdf/

        Parameters
        ----------
        tags

        Returns
        -------

        """
        l = []
        self.console.print('=========>start converting bam to df...')
        stime = time.time()
        for id, read in enumerate(self.pysam_bam):
            # print(id)

            read_tags = read.get_tags()
            # print(read_tags)
            rt_dict = {k: v for k, v in read_tags}
            rt_keys = [*rt_dict.keys()]
            ### @@ read_tags | per read
            ### @@ rt_dict | per read
            ### @@ rt_keys | per read
            # 1st read
            # [('XA', 2), ('MD', '44T2T5'), ('NM', 2)]
            # {'XA': 2, 'MD': '44T2T5', 'NM': 2}
            # ['XA', 'MD', 'NM']
            # 2nd read
            # [('XA', 2), ('MD', '12A0C6'), ('NM', 2)]
            # {'XA': 2, 'MD': '12A0C6', 'NM': 2}
            # ['XA', 'MD', 'NM']
            # ...
            # the last read
            # [('XA', 2), ('MD', '1T4A13'), ('NM', 2)]
            # {'XA': 2, 'MD': '1T4A13', 'NM': 2}
            # ['XA', 'MD', 'NM']
            tag_keys = [rt_dict[k] if k in rt_keys else 'None' for k in tags]
            ### @@ tag_keys
            # [2, '44T2T5', 2]
            # [2, '12A0C6', 2]
            # ...
            # [2, '1T4A13', 2]
            vignette = [
                id,
                read.query_name,
                read.flag,
                read.reference_id,
                read.pos,
                read.mapping_quality,
                read.cigar,
                read.query_sequence,
                read.next_reference_id,
                read.next_reference_start,
                read.query_qualities,
                # read.template_length,
                read,
            ] + tag_keys
            l.append(vignette)
        df = pd.DataFrame(
            l,
            columns=[
                'id',
                'query_name',
                'flag',
                'reference_id',
                'genome_pos',
                'mapping_quality',
                'cigar',
                'query_sequence',
                'next_reference_id',
                'next_reference_start',
                'query_qualities',
                # 'template_length',
                'read',
            ] + tags,
        )
        if filter_by == 'sequence':
            df = df.loc[~df['query_sequence'].isin([None])] # .reset_index(drop=True)
            # df['query_sequence'] = df['query_sequence'].apply(lambda x: x[:20])
            return df
        else:
            # print(df['XA'].loc[df['reference_id'] != -1].shape)
            # print(df['MD'].loc[df['MD'] != 'None'].shape)
            # print(df['NM'].loc[df['NM'] != 'None'].shape)
            # print(df['XS'].loc[df['XS'] != 'None'].shape)
            # print(df['XT'].loc[df['XT'] != 'None'].shape)
            self.console.print('=========>time to df: {:.3f}s'.format(time.time() - stime))
            # df = df.sample(
            #     n=50,
            #     replace=True,
            # )
            return df

    def todf_trust4(
            self,
            tags=[],
            filter_by=None,
    ):
        l = []
        self.console.print('=========>start converting bam to df...')
        stime = time.time()
        for id, read in enumerate(self.pysam_bam):
            # print(id)
            # print(read.get_tag('CB'))
            # arr_query_name = read.query_name.split('_')
            # BC = arr_query_name[1]
            # UMI = arr_query_name[2]
            # read.set_tag('CB', BC)
            # read.set_tag('UB', UMI)
            # read.query_name = arr_query_name[0]
            # print(read.query_name)

            # if read.query_sequence != None:
            # read.query_sequence = None
            # len(read.query_sequence)

            # print(len(read.query_sequence))
            # print(read.query_qualities[:20])
            # t = read.query_qualities[:20]
            # u = read.query_sequence[:20]
            # print(u)
            # read.query_sequence = u
            # read.query_qualities = None

            read_tags = read.get_tags()
            # print(read_tags)
            rt_dict = {k: v for k, v in read_tags}
            rt_keys = [*rt_dict.keys()]
            ### @@ read_tags | per read
            ### @@ rt_dict | per read
            ### @@ rt_keys | per read
            # 1st read
            # [('XA', 2), ('MD', '44T2T5'), ('NM', 2)]
            # {'XA': 2, 'MD': '44T2T5', 'NM': 2}
            # ['XA', 'MD', 'NM']
            # 2nd read
            # [('XA', 2), ('MD', '12A0C6'), ('NM', 2)]
            # {'XA': 2, 'MD': '12A0C6', 'NM': 2}
            # ['XA', 'MD', 'NM']
            # ...
            # the last read
            # [('XA', 2), ('MD', '1T4A13'), ('NM', 2)]
            # {'XA': 2, 'MD': '1T4A13', 'NM': 2}
            # ['XA', 'MD', 'NM']
            tag_keys = [rt_dict[k] if k in rt_keys else 'None' for k in tags]
            ### @@ tag_keys
            # [2, '44T2T5', 2]
            # [2, '12A0C6', 2]
            # ...
            # [2, '1T4A13', 2]
            vignette = [
                id,
                read.query_name,
                read.flag,
                read.reference_id,
                read.pos,
                read.mapping_quality,
                read.cigar,
                read.query_sequence,
                read.next_reference_id,
                read.next_reference_start,
                read.query_qualities,
                # read.template_length,
                read,
            ] + tag_keys
            l.append(vignette)
        df = pd.DataFrame(
            l,
            columns=[
                'id',
                'query_name',
                'flag',
                'reference_id',
                'genome_pos',
                'mapping_quality',
                'cigar',
                'query_sequence',
                'next_reference_id',
                'next_reference_start',
                'query_qualities',
                # 'template_length',
                'read',
            ] + tags,
        )
        if filter_by == 'sequence':
            df = df.loc[~df['query_sequence'].isin([None])] # .reset_index(drop=True)
            # df['query_sequence'] = df['query_sequence'].apply(lambda x: x[:20])
            return df
        else:
            # print(df['XA'].loc[df['reference_id'] != -1].shape)
            # print(df['MD'].loc[df['MD'] != 'None'].shape)
            # print(df['NM'].loc[df['NM'] != 'None'].shape)
            # print(df['XS'].loc[df['XS'] != 'None'].shape)
            # print(df['XT'].loc[df['XT'] != 'None'].shape)
            self.console.print('=========>time to df: {:.3f}s'.format(time.time() - stime))
            # df = df.sample(
            #     n=50,
            #     replace=True,
            # )
            return df

    def todf11_depr(self, ):
        """
        Note
        ----
        Deprecated.

        Returns
        -------
        Dataframe of a bam file

        """
        l = []
        self.console.print('=========>start converting bam to df')
        import time
        stime = time.time()
        for id, read in enumerate(self.pysam_bam):
            read_piece = {
                'id': id,
                'query_name': read.query_name,
                'flag': read.flag,
                'reference_id': read.reference_id,
                'reference_start': read.reference_start,
                'mapping_quality': read.mapping_quality,
                'cigar': read.cigar,
                'query_sequence': read.query_sequence,
                'next_reference_id': read.next_reference_id,
                'next_reference_start': read.next_reference_start,
                'template_length': read.template_length,
                'query_qualities': read.query_qualities,
            }
            l.append(read_piece)
        df = pd.DataFrame.from_dict(l)
        self.console.print('=========>time to df: {:.3f}s'.format(time.time() - stime))
        return df


if __name__ == "__main__":
    from umiche.path import to

    umiche = Reader(
        # bam_fpn=to('data/example.bam'),
        # bam_fpn=to('data/example_bundle1.bam'),
        # bam_fpn=to('data/hgmm_100_STAR_FC_sorted.bam'),
        # to('example/data/assigned_sorted_dedup.bam'),
        # bam_fpn=to('data/simu/monomer/sc/seq_err/permute_0/trimmed/seq_err_0.bam'),
        # bam_fpn=to('data/simu/umi/seq_errs/trimer/permute_0/bam/bipartite/seq_err_0.bam'),
        # bam_fpn=to('data/simu/trimer/pcr8/seq_errs/permute_0/trimmed/seq_err_18.bam'),
        # bam_fpn=to('example/data/deduplicated.bam'),
        # bam_fpn=to('example/data/RM82CLK1_S3_featurecounts_gene_sorted.bam'),

        # bam_fpn=to('data/bone_marrow/example.bam')
        # bam_fpn=to('data/bone_marrow/stripped_XT_sorted.bam')
        # bam_fpn=to('data/bone_marrow/stripped_XT_sorted_little.bam')
        # bam_fpn=to('data/bone_marrow/merge_corrected_sorted_little.bam')
        # bam_fpn=to('data/bone_marrow/merge_corrected_sorted.bam')
        # bam_fpn=to('data/bone_marrow/little.bam')
        # bam_fpn=to('data/bone_marrow/trimmed1.bam')
        bam_fpn=to('data/bone_marrow/trimmed2.bam')
    )

    # df = umiche.todf(tags=[])
    # df = umiche.todf(tags=['PO'])
    # df = umiche.todf_trust4(tags=[], filter_by='sequence')
    # print(df)
    # print(df.loc[0].values)

    df = umiche.todf_trust4(tags=['CB', 'UB'], filter_by='sequence')
    df['r1'] = df.apply(lambda x: x['CB'] + x['UB'], axis=1)
    print(df[['r1', 'query_name']])

    from umiche.fastq.Writer import Writer as fastqwriter
    fastqwriter().togz(
        list_2d=df[['r1', 'query_name']].values.tolist(),
        sv_fpn=to('data/bone_marrow/trimmed2_r1.fastq.gz'),
    )

    # from umiche.bam.Writer import Writer as aliwriter
    # aliwriter(df=df).tobam(
    #     tobam_fpn=to('data/bone_marrow/trimmed2.bam'),
    #     tmpl_bam_fpn=to('data/bone_marrow/merge_corrected_sorted_little.bam'),
    #     whitelist=df.index,
    # )
    # print(df.columns)
    # print(df.query_name)
    # df = umiche.todf(tags=['BC', 'XS', 'XT'])
    # df = umiche.todf(tags=['XS', 'XT'])
    # print(df[['query_name', 'BC', 'XS', 'XT']])

    # df = df.loc[df['XS'] == 'Assigned']
    # print(df)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import scipy.optimize as opt


class Anchor:

    def __init__(
            self,
            quant_captured,
            quant_anchor_captured,
            criteria,
            condition,
    ):
        self.quant_captured = quant_captured
        self.quant_anchor_captured = quant_anchor_captured

        self.df_quant_captured = pd.DataFrame.from_dict(self.quant_captured)
        self.df_quant_anchor_captured = pd.DataFrame.from_dict(self.quant_anchor_captured)

        self.df_quant_captured_mean = self.df_quant_captured.mean(axis=1).values
        self.df_quant_anchor_captured_mean = self.df_quant_anchor_captured.mean(axis=1).values

        self.err_captured = self.df_quant_captured.max(axis=1).values - self.df_quant_captured_mean
        self.err_anchor_captured = self.df_quant_anchor_captured.max(axis=1).values - self.df_quant_anchor_captured_mean

        # self.err_captured = pd.DataFrame.from_dict(quant_captured).std(axis=1).values
        # self.err_anchor_captured = pd.DataFrame.from_dict(quant_anchor_captured).std(axis=1).values

        self.criteria = criteria
        self.condition = condition

        self.palette = sns.color_palette("Set3")[3:5][::-1]

        self.meas = {
            'without anchor': self.df_quant_captured_mean,
            'with anchor (BAGC)': self.df_quant_anchor_captured_mean,
        }

        self.meas_errs = {
            'without anchor': self.err_captured,
            'with anchor (BAGC)': self.err_anchor_captured,
        }

        sns.set(font="Helvetica")
        sns.set_style("ticks")

    def fit_logistic(self, x, a, b, c, d):
        return a / (1. + np.exp(-c * (x - d))) + b

    def draw_line_broken(
            self,
            notion=False,
    ):
        x_notions = [str(criterion) for criterion in self.criteria]
        x = np.arange(len(x_notions))

        fig, ax = plt.subplots(nrows=2, ncols=1, figsize=(6, 5), sharex=True)
        for i in range(2):
            for j, (attr, measurement) in enumerate(self.meas.items()):
                # ax[i].scatter(
                #     x,
                #     measurement,
                #     color='grey',
                #     linewidths=2,
                #     s=60,
                #     facecolors='none',
                # )
                ax[i].errorbar(
                    x,
                    measurement,
                    yerr=self.meas_errs[attr],
                    elinewidth=2,
                    markersize=5,
                    fmt='o',
                    markerfacecolor='none',
                )

                def f(x, a, b, c, d):
                    return a / (1. + np.exp(-c * (x - d))) + b

                popt, pcov = opt.curve_fit(f, x, measurement, method="trf")
                print(popt)
                y_fit = f(x, *popt)

                # new_x = np.linspace(0, len(x_notions), 100)
                # import scipy.interpolate as spi
                # ipo3 = spi.splrep(x, measurement, k=3)+
                # iy3 = spi.splev(new_x, ipo3)
                ax[i].plot(
                    x,
                    y_fit,
                    # new_x,
                    # iy3,
                    color=self.palette[j],
                    label=attr,
                    linewidth=3,
                )
                ax[i].spines['right'].set_color('none')
                ax[i].spines['top'].set_color('none')
                ax[i].spines['bottom'].set_linewidth(2)
                ax[i].spines['left'].set_linewidth(2)

        ax[0].spines.bottom.set_visible(False)
        ax[0].xaxis.set_ticks_position('none')
        ax[1].xaxis.tick_bottom()
        ax[0].tick_params(axis='y', labelsize=14, width=2)
        ax[1].tick_params(axis='y', labelsize=14, width=2)

        if notion:
            # a / (1. + np.exp(-c * (x - d))) + b
            # import matplotlib.patches as mpatches
            ax[0].annotate(
                r'$\frac{-0.241}{1+e^{-0.709(x-11.128)}}$',
                # xy=(4, 0.9),
                # xytext=(0.38, 0.5),
                xy=(6, 0.89),
                xytext=(0.38, 0.5),
                textcoords='axes fraction',
                arrowprops=dict(
                    arrowstyle="fancy",
                    color="0.5",
                    # patchB=mpatches.Ellipse((5, 0.23), 0.3, 0.4, angle=30, alpha=0.2),
                    facecolor='dimgrey',
                    shrinkB=5,
                    connectionstyle="arc3,rad=0.3",
                ),
                fontsize=16,
                horizontalalignment='right', verticalalignment='top',
            )
            ax[1].annotate(
                r'$\frac{-0.364}{1+e^{-0.835(x-14.448)}}$',
                xy=(5, 0.23),
                xytext=(0.78, 1.2),
                textcoords='axes fraction',
                arrowprops=dict(
                    arrowstyle="fancy",
                    color="0.5",
                    # patchB=mpatches.Ellipse((5, 0.23), 0.3, 0.4, angle=30, alpha=0.2),
                    facecolor='dimgrey',
                    shrinkB=5,
                    connectionstyle="arc3,rad=0.3",
                ),
                fontsize=16,
                horizontalalignment='right', verticalalignment='top',
            )

        ax[1].set_xlabel('Insertion error rate', fontsize=18)

        ax[0].set_ylim(.59, 0.95)
        ax[1].set_ylim(-0.06, .25)

        d = .5  # proportion of vertical to horizontal extent of the slanted line
        kwargs = dict(marker=[(-1, -d), (1, d)], markersize=12,
                      linestyle="none", color='k', mec='k', mew=1, clip_on=False)
        ax[0].plot([0], [0], transform=ax[0].transAxes, **kwargs)
        ax[1].plot([0], [1], transform=ax[1].transAxes, **kwargs)

        def formatter(x, pos):
            del pos
            return '{}'.format(int(x * 100))

        ax[0].yaxis.set_major_formatter(formatter)
        ax[1].yaxis.set_major_formatter(formatter)

        # ax[1].arrow(2, 0.05, 4, 4, width=0.2)

        ax[1].set_xticks(np.linspace(0, 16, 5))
        # for i in [1e-05, 1e-04, 1e-03, 1e-02, 1e-01]
        ax[1].set_xticklabels(["$10^{-5}$", "$10^{-4}$", "$10^{-3}$", "$10^{-2}$", "$10^{-1}$"], fontsize=14)

        fig.text(
            0.04,
            0.5,
            'Percentage of reads captured (%)',
            va='center',
            rotation='vertical',
            fontsize=18,
        )

        plt.legend(ncol=1, fontsize=16)

        plt.subplots_adjust(
            top=0.94,
            bottom=0.129,
            left=0.14,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        # plt.savefig('./data/Figure4d.eps', format='eps')
        plt.show()

    def draw(
            self,
            notion=False,
            xy=None,
            xytext=None,
    ):
        x_notions = [str(criterion) for criterion in self.criteria]
        x = np.arange(len(x_notions))
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 5), sharex=True)

        for j, (attr, measurement) in enumerate(self.meas.items()):
            # ax[i].scatter(
            #     x,
            #     measurement,
            #     color='grey',
            #     linewidths=2,
            #     s=60,
            #     facecolors='none',
            # )
            ax.errorbar(
                x,
                measurement,
                yerr=self.meas_errs[attr],
                elinewidth=2,
                markersize=5,
                fmt='o',
                markerfacecolor='none',
            )

            popt, pcov = opt.curve_fit(self.fit_logistic, x, measurement, method="trf")
            y_fit = self.fit_logistic(x, *popt)

            # new_x = np.linspace(0, len(x_notions), 100)
            # import scipy.interpolate as spi
            # ipo3 = spi.splrep(x, measurement, k=3)+
            # iy3 = spi.splev(new_x, ipo3)  # 根据观测点和样条参数，生成插值
            ax.plot(
                x,
                y_fit,
                # new_x,
                # iy3,
                color=self.palette[j],
                label=attr,
                linewidth=3,
            )
            ax.spines['right'].set_color('none')
            ax.spines['top'].set_color('none')
            ax.spines['bottom'].set_linewidth(2)
            ax.spines['left'].set_linewidth(2)

        ax.xaxis.set_ticks_position('none')
        ax.xaxis.tick_bottom()
        ax.tick_params(axis='y', labelsize=14, width=2)
        ax.set_xlabel(self.condition, fontsize=18)
        ax.set_ylim(-0.06, 1.1)
        ax.set_xticks(np.linspace(0, 16, 5))
        ax.set_xticklabels(["$10^{-5}$", "$10^{-4}$", "$10^{-3}$", "$10^{-2}$", "$10^{-1}$"], fontsize=14)

        if notion:
            # a / (1. + np.exp(-c * (x - d))) + b
            # import matplotlib.patches as mpatches
            ax.annotate(
                # pcr_ins
                # r'$\frac{-0.984}{1+e^{-0.834(x-8.747)}}$',
                # xy=(5.5, 0.88),
                # xytext=(0.38, 0.65),

                # # seq_ins
                # r'$\frac{-1.064}{1+e^{-0.706(x-11.176)}}$',
                # xy=(7, 0.92),
                # xytext=(0.48, 0.7),

                # seq_del
                # r'$\frac{-1.064}{1+e^{-0.704(x-11.165)}}$',
                # xy=(7, 0.92),
                # xytext=(0.48, 0.7),

                # pcr_del
                # r'$\frac{-0.979}{1+e^{-0.865(x-8.742)}}$',
                # xy=(5.5, 0.88),
                # xytext=(0.38, 0.65),

                r'$\frac{-0.979}{1+e^{-0.865(x-8.742)}}$',
                xy=xy,
                xytext=xytext,
                textcoords='axes fraction',
                arrowprops=dict(
                    arrowstyle="fancy",
                    color="0.5",
                    # patchB=mpatches.Ellipse((5, 0.23), 0.3, 0.4, angle=30, alpha=0.2),
                    facecolor='dimgrey',
                    shrinkB=5,
                    connectionstyle="arc3,rad=-0.3",
                ),
                fontsize=18,
                horizontalalignment='right', verticalalignment='top',
            )
            ax.annotate(
                # pcr_ins
                # r'$\frac{-0.689}{1+e^{-0.916(x-13.732)}}$',
                # xy=(14.2, 0.55),
                # xytext=(0.95, 0.35),

                # seq_ins
                # r'$\frac{-0.293}{1+e^{-0.919(x-14.337)}}$',
                # xy=(15.2, 0.8),
                # xytext=(1.0, 0.6),

                # seq_del
                # r'$\frac{-0.381}{1+e^{-0.895(x-14.318)}}$',
                # xy=(15, 0.75),
                # xytext=(1.0, 0.6),

                # pcr_del
                # r'$\frac{-0.832}{1+e^{-0.834(x-13.753)}}$',
                # xy=(14., 0.5),
                # xytext=(0.95, 0.35),

                r'$\frac{-0.832}{1+e^{-0.834(x-13.753)}}$',
                xy=xy,
                xytext=xytext,
                textcoords='axes fraction',
                arrowprops=dict(
                    arrowstyle="fancy",
                    color="0.5",
                    # patchB=mpatches.Ellipse((5, 0.23), 0.3, 0.4, angle=30, alpha=0.2),
                    facecolor='dimgrey',
                    shrinkB=5,
                    connectionstyle="arc3,rad=-0.3",
                ),
                fontsize=18,
                horizontalalignment='right', verticalalignment='top',
            )

        def formatter(x, pos):
            del pos
            return '{}'.format(int(x * 100))

        ax.yaxis.set_major_formatter(formatter)
        ax.yaxis.set_major_formatter(formatter)

        fig.text(
            0.04,
            0.5,
            'Percentage of reads recovered (%)',
            va='center',
            rotation='vertical',
            fontsize=18,
        )
        plt.legend(ncol=1, fontsize=16)
        plt.subplots_adjust(
            top=0.94,
            bottom=0.129,
            left=0.14,
            right=0.98,
            # hspace=0.40,
            # wspace=0.15,
        )
        # plt.savefig('../data/' + x_label + '.eps', format='eps')
        plt.show()

    def simple(self, ):
        x_notions = [str(criterion) for criterion in self.criteria]
        x = np.arange(len(x_notions))
        width = 0.35
        multiplier = 0

        fig, ax = plt.subplots(layout='constrained', figsize=(10, 6))

        for attr, measurement in self.meas.items():
            print(attr, measurement)
            offset = width * multiplier
            rects = ax.bar(x + offset, measurement, width, label=attr)
            ax.bar_label(rects, fmt="{:.1%}", padding=2, fontsize=10)
            multiplier += 1

        ax.set_ylabel('Successfully captured reads', fontsize=16)
        ax.set_title('scCOLOR-seq v2 bead design simulation', fontsize=14)
        ax.set_xticks(x + width, x_notions, ha='right', rotation=20)
        ax.set_xlabel(self.condition, fontsize=16)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)

        ax.legend(ncols=2)
        ax.set_ylim(0, 1)

        plt.show()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import numpy as np


class Number:

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def uniform(self, low, high, num, use_seed=True, seed=1):
        """

        Parameters
        ----------
        low
        high
        num
        use_seed
        seed

        Returns
        -------

        """
        if use_seed:
            state = np.random.RandomState(seed)
            return state.randint(
                low=low,
                high=high,
                size=num
            )
        else:
            return np.random.randint(
                low=low,
                high=high,
                size=num
            )


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import warnings

import numpy as np

warnings.filterwarnings("ignore")

import json
import pandas as pd
from umiche.simu.Parameter import Parameter as params
from umiche.trim.Template import Template as trimmer
from umiche.fastq.Convert import Convert as fastqconverter

from umiche.deduplicate.OnePos import OnePos as dedupop
from umiche.plot.scenario.TraceSingle import TraceSingle as plothetero

from umiche.deduplicate.heterogeneity.Trace import Trace as umitrace
from umiche.bam.Relation import Relation as umirel

from umiche.util.Writer import Writer as fwriter
from umiche.util.Console import Console


class Heterogeneity:

    def __init__(
            self,
            scenario,
            method,
            param_fpn=None,
            is_trim=False,
            is_tobam=False,
            is_dedup=False,
            is_sv=False,
            verbose=False,
            **kwargs,
    ):
        self.scenario = scenario
        self.method = method
        self.is_sv = is_sv
        self.kwargs = kwargs
        self.verbose = verbose

        self.params = params(param_fpn=param_fpn)
        self.fwriter = fwriter()

        self.console = Console()
        self.console.verbose = self.verbose

        columns = ['diff_origin', 'same_origin', 'total', 'scenario', 'method', 'permutation']
        self.df_apv_cnt = pd.DataFrame(columns=columns)
        self.df_disapv_cnt = pd.DataFrame(columns=columns)
        self.df_apv_pct = pd.DataFrame(columns=columns)
        self.df_disapv_pct = pd.DataFrame(columns=columns)
        self.df_dedup = pd.DataFrame()
        self.node_repr_dict = {}
        print("===>dedup method: {}".format(self.method))
        for perm_num_i in range(self.params.fixed['permutation_num']):
            print("===>No.{} permutation".format(perm_num_i))
            self.console.print("===>permutation number {}".format(perm_num_i))
            dedup_arr = []
            self.node_repr_dict[perm_num_i] = {}
            for id, scenario_i in enumerate(self.params.varied[self.scenario]):
                if self.scenario == 'pcr_nums':
                    self.fn_mark = str(id) # scenario_i
                elif self.scenario == 'pcr_errs':
                    self.fn_mark = str(id)
                elif self.scenario == 'seq_errs':
                    self.fn_mark = str(id)
                elif self.scenario == 'ampl_rates':
                    self.fn_mark = str(id)
                elif self.scenario == 'seq_deps':
                    self.fn_mark = str(id)
                elif self.scenario == 'umi_lens':
                    self.fn_mark = str(id)
                elif self.scenario == 'umi_nums':
                    self.fn_mark = str(id)
                else:
                    self.fn_mark = str(scenario_i)
                self.fn_prefix = self.params.file_names[self.scenario] + self.fn_mark
                self.fastq_location = self.params.work_dir + self.scenario + '/permute_' + str(perm_num_i) + '/'
                if is_trim:
                    if self.scenario == 'umi_lens':
                        self.params.trimmed['umi_1']['len'] = scenario_i
                    self.console.print("======>fastq is being trimmed.")
                    self.params.trimmed['fastq']['fpn'] = self.fastq_location + self.fn_prefix + '.fastq.gz'
                    self.params.trimmed['fastq']['trimmed_fpn'] = self.fastq_location + 'trimmed/' + self.fn_prefix + '.fastq.gz'
                    umitrim_parser = trimmer(params=self.params.trimmed, verbose=self.verbose)
                    df = umitrim_parser.todf()
                    umitrim_parser.togz(df)
                if is_tobam:
                    self.console.print("======>fastq converter to bam is being used.")
                    fastqconverter(
                        fastq_fpn=self.fastq_location + 'trimmed/' + self.fn_prefix + '.fastq.gz',
                        bam_fpn=self.fastq_location + 'trimmed/bam/' + self.fn_prefix + '.bam',
                    ).tobam()
                if is_dedup:
                    self.console.print("======>reads are being deduplicated.")
                    # neither nor
                    if not (isinstance(self.params.dedup['inflat_val'], list) or isinstance(self.params.dedup['exp_val'], list)):
                        self.console.print("======>the working mode is not to optimise MCL inflation and expasion")
                        optim_arr = [self.params.dedup['inflat_val']]
                        inflat_flag = True
                    else:
                        self.console.print("======>the working mode is to optimise MCL inflation")
                        if isinstance(self.params.dedup['inflat_val'], list):
                            optim_arr = self.params.dedup['inflat_val']
                            inflat_flag = True
                        else:
                            self.console.print("======>the working mode is to optimise MCL expasion")
                            optim_arr = self.params.dedup['exp_val']
                            inflat_flag = False

                    for optim_val in optim_arr:
                        dedup_ob = dedupop(
                            # bam_fpn=self.fastq_location + 'trimmed/bam/pcr_num_12.bam',
                            # bam_fpn=self.fastq_location + 'trimmed/bam/pcr_err_10.bam',
                            # bam_fpn=self.fastq_location + 'trimmed/bam/seq_err_12.bam',
                            # bam_fpn=self.fastq_location + 'trimmed/bam/ampl_rate_8.bam',
                            # bam_fpn=self.fastq_location + 'trimmed/bam/umi_len_6.bam',
                            # bam_fpn=self.fastq_location + 'trimmed/bam/seq_dep_7.bam',
                            bam_fpn=self.fastq_location + 'trimmed/bam/' + self.fn_prefix + '.bam',
                            mcl_fold_thres=self.params.dedup['mcl_fold_thres'],
                            inflat_val=optim_val if inflat_flag else self.params.dedup['inflat_val'],
                            exp_val=optim_val if not inflat_flag else self.params.dedup['exp_val'],
                            # inflat_val=self.params.dedup['inflat_val'],
                            # exp_val=self.params.dedup['exp_val'],
                            iter_num=self.params.dedup['iter_num'],
                            ed_thres=self.params.dedup['ed_thres'],
                            dbscan_eps=self.params.dedup['dbscan_eps'],
                            dbscan_min_spl=self.params.dedup['dbscan_min_spl'],
                            birch_thres=self.params.dedup['birch_thres'],
                            birch_n_clusters=self.params.dedup['birch_n_clusters'],
                            work_dir=self.params.work_dir,
                            heterogeneity=True,
                            verbose=False,
                            **self.kwargs
                        )
                        df = self.tool(dedup_ob)[self.method]()
                        if self.method == "unique":
                            dedup_arr.append(df.num_uniq_umis.values[0])
                            print("No.{}->{} for {} dedup cnt: {}".format(
                                id,
                                scenario_i,
                                self.scenario,
                                df.num_uniq_umis.values[0]),
                            )
                        else:
                            dedup_arr.append(df.dedup_cnt.values[0])
                            print("No.{}->{} for {} dedup cnt: {}".format(
                                id,
                                scenario_i,
                                self.scenario,
                                df.dedup_cnt.values[0]),
                            )
                    # print(df.apv.values[0])
                    # print(len(df[self.method + '_repr_nodes'].loc[1]))

                        if self.method not in ['adjacency', 'cluster', 'unique']:
                            umiold = umirel(
                                df=dedup_ob.df_bam,
                                verbose=self.verbose,
                            )
                            umiidtrace = umitrace(
                                df_umi_uniq_val_cnt=umiold.df_umi_uniq_val_cnt,
                                umi_id_to_origin_id_dict=umiold.umi_id_to_origin_id_dict,
                            )
                            series_2d_arr_apv, series_2d_arr_disapv = umiidtrace.format(method=self.method, df=df)

                            series_dict_origin_apv = umiidtrace.match_representative(series_2d_arr=series_2d_arr_apv)
                            self.node_repr_dict[perm_num_i][scenario_i] = series_dict_origin_apv.to_dict()
                            # print(series_2d_arr_disapv)
                            if not series_2d_arr_apv.empty:
                                apv_cnt_dict = umiidtrace.edge_class(series_2d_arr=series_2d_arr_apv, sort='cnt')
                                # print(apv_cnt_dict)
                                apv_pct_dict = umiidtrace.edge_class(series_2d_arr=series_2d_arr_apv, sort='pct')
                                apv_cnt_dict['permutation'] = perm_num_i
                                apv_cnt_dict['method'] = self.method
                                apv_cnt_dict['scenario'] = scenario_i
                                apv_pct_dict['permutation'] = perm_num_i
                                apv_pct_dict['method'] = self.method
                                apv_pct_dict['scenario'] = scenario_i
                                self.df_apv_cnt = pd.concat([self.df_apv_cnt, pd.DataFrame.from_dict(apv_cnt_dict, orient='index').T]).reset_index(drop=True)
                                self.df_apv_pct = pd.concat([self.df_apv_pct, pd.DataFrame.from_dict(apv_pct_dict, orient='index').T]).reset_index(drop=True)
                            if not series_2d_arr_disapv.empty:
                                disapv_cnt_dict = umiidtrace.edge_class(series_2d_arr=series_2d_arr_disapv, sort='cnt')
                                disapv_pct_dict = umiidtrace.edge_class(series_2d_arr=series_2d_arr_disapv, sort='pct')
                                disapv_cnt_dict['permutation'] = perm_num_i
                                disapv_cnt_dict['method'] = self.method
                                disapv_cnt_dict['scenario'] = scenario_i
                                disapv_pct_dict['permutation'] = perm_num_i
                                disapv_pct_dict['method'] = self.method
                                disapv_pct_dict['scenario'] = scenario_i
                                self.df_disapv_cnt = pd.concat([self.df_disapv_cnt, pd.DataFrame.from_dict(disapv_cnt_dict, orient='index').T]).reset_index(drop=True)
                                self.df_disapv_pct = pd.concat([self.df_disapv_pct, pd.DataFrame.from_dict(disapv_pct_dict, orient='index').T]).reset_index(drop=True)

                            # print(self.df_apv_pct)
                            # plothetero(
                            #     df_apv=self.df_apv_cnt,
                            #     df_disapv=self.df_disapv_cnt,
                            # ).line_apv_disapv()
            self.df_dedup['pn' + str(perm_num_i)] = dedup_arr
            print(self.df_dedup)

        if self.is_sv:
            sv_dedup_fpn = self.params.work_dir + '/' + scenario + '/' + str(self.method) + '_dedup' + '.txt'
            sv_apv_cnt_fpn = self.params.work_dir + '/' + scenario + '/' + str(self.method) + '_apv_cnt' + '.txt'
            sv_disapv_cnt_fpn = self.params.work_dir + '/' + scenario + '/' + str(self.method) + '_disapv_cnt' + '.txt'
            sv_apv_pct_fpn = self.params.work_dir + '/' + scenario + '/' + str(self.method) + '_apv_pct' + '.txt'
            sv_disapv_pct_fpn = self.params.work_dir + '/' + scenario + '/' + str(self.method) + '_disapv_pct' + '.txt'
            sv_node_repr_fpn = self.params.work_dir + '/' + scenario + '/' + str(self.method) + '_node_repr' + '.json'
            self.fwriter.generic(df=self.df_dedup, sv_fpn=sv_dedup_fpn, header=True, )
            self.fwriter.generic(df=self.df_apv_cnt, sv_fpn=sv_apv_cnt_fpn, header=True, )
            self.fwriter.generic(df=self.df_disapv_cnt, sv_fpn=sv_disapv_cnt_fpn, header=True, )
            self.fwriter.generic(df=self.df_apv_pct, sv_fpn=sv_apv_pct_fpn, header=True, )
            self.fwriter.generic(df=self.df_disapv_pct, sv_fpn=sv_disapv_pct_fpn, header=True, )
            with open(sv_node_repr_fpn, 'w') as f:
                json.dump(self.node_repr_dict, f)

    def tool(self, dedup_ob):
        return {
            'unique': dedup_ob.unique,
            'cluster': dedup_ob.cluster,
            'adjacency': dedup_ob.adjacency,
            'directional': dedup_ob.directional,
            'mcl': dedup_ob.mcl,
            'mcl_val': dedup_ob.mcl_val,
            'mcl_ed': dedup_ob.mcl_ed,
            'mcl_cc_all_node_umis': dedup_ob.mcl_cc_all_node_umis,
            'dbscan_seq_onehot': dedup_ob.dbscan_seq_onehot,
            'birch_seq_onehot': dedup_ob.birch_seq_onehot,
            'hdbscan_seq_onehot': dedup_ob.hdbscan_seq_onehot,
            'aprop_seq_onehot': dedup_ob.aprop_seq_onehot,
            # 'set_cover': dedup_ob.set_cover,
        }

    @property
    def df_stat(self, ):
        return {
            i: pd.DataFrame(columns=['diff_origin', 'same_origin', 'total']) for i in [
                'df_apv_cnt',
                'df_disapv_cnt',
                'df_disapv_pct',
                'df_disapv_pct',
            ]}


if __name__ == "__main__":
    from umiche.path import to

    p = Heterogeneity(
        # scenario='pcr_nums',
        # scenario='pcr_errs',
        # scenario='seq_errs',
        # scenario='ampl_rates',
        # scenario='umi_lens',
        scenario='seq_deps',
        # scenario='umi_nums',

        # method='unique',
        # method='cluster',
        # method='adjacency',
        # method='directional',
        # method='mcl',
        # method='mcl_val',
        # method='mcl_ed',
        # method='mcl_cc_all_node_umis',
        # method='dbscan_seq_onehot',
        method='birch_seq_onehot',
        # method='aprop_seq_onehot',
        # method='hdbscan_seq_onehot',
        # method='set_cover',

        # is_trim=True,
        # is_tobam=False,
        # is_dedup=False,

        # is_trim=False,
        # is_tobam=True,
        # is_dedup=False,

        is_trim=False,
        is_tobam=False,
        is_dedup=True,

        is_sv=True,

        param_fpn=to('data/params.yml'),
    )


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"



class ReformKit:

    def __init__(self, ):
        pass

    def maxid(
            self,
            df_umi_uniq_val_cnt,
            cc,
    ):
        """

        Parameters
        ----------
        df_umi_uniq_val_cnt
        cc

        Returns
        -------

        """
        xx = [*df_umi_uniq_val_cnt.loc[df_umi_uniq_val_cnt.index.isin(cc)].sort_values(ascending=False).to_dict().keys()]
        return xx[0]

    def sort_vals(
            self,
            df_umi_uniq_val_cnt,
            cc,
    ):
        """

        Parameters
        ----------
        df_umi_uniq_val_cnt
        cc

        Returns
        -------

        """
        return df_umi_uniq_val_cnt.loc[df_umi_uniq_val_cnt.index.isin(cc)].sort_values(ascending=False).to_dict()

    def breakpoint(self, x, connected_components):
        print('MCL clusters:\n {}'.format(x['clusters']))
        print('CCS:\n {}'.format(len(connected_components)))
        print('CC:\n {}'.format(x['cc_vertices']))
        print('Edge list of MCL clusters:\n {}'.format(x['edge_list']))
        print('Graph adj list of MCL clusters:\n {}'.format(x['graph_cc_adj']))
        print('\n')
        # df.to_csv('./sd.txt', sep='\t', index=False, header=True)
        return

    def keymap(
            self,
            graph_adj,
            reverse=False,
    ):
        """

        Parameters
        ----------
        graph_adj
        reverse

        Returns
        -------

        """
        keys = [*graph_adj.keys()]
        glen = len(keys)
        if reverse:
            return {k: keys[k] for k in range(glen)}
        else:
            return {keys[k]: k for k in range(glen)}

    def graph_cc_adj(self, cc, graph_adj):
        """

        Parameters
        ----------
        cc
            The first parameter.
        graph_adj
            The se parameter.

        Returns
        -------

        """
        return {node: graph_adj[node] for node in cc}

    def key2node(self, list_2d, keymap):
        """

        Parameters
        ----------
        list_2d
        keymap

        Returns
        -------

        """
        return [[keymap[i] for i in lis] for lis in list_2d]

    def onehot(
            self,
            umi,
    ):
        import numpy as np
        from umiche.util.Single import Single as dnasgl
        nt_to_int_dict = dnasgl().todict(nucleotides=dnasgl().get(universal=True), reverse=False)
        # int_to_nt_dict = dnasgl().todict(nucleotides=dnasgl().get(universal=True), reverse=True)
        umi_ltr2digi = [nt_to_int_dict[i] for i in list(umi)]
        ids_first_pos = [i*4 for i in range(len(umi))]
        ids_to_be_one = [i+j for i, j in zip(umi_ltr2digi, ids_first_pos)]
        # print(ids_to_be_one)
        one_hot = np.zeros(len(umi)*4)
        one_hot[ids_to_be_one] = 1
        return one_hot.astype(int)



import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE


class dimensionReduction:

    def __init__(self, ):
        pass

    def single(self, X, y, tech='TSNE', marker_size=2, cmap='tab20b', title=''):
        """https://matplotlib.org/stable/tutorials/colors/colormaps.html"""
        sns.set(font="Helvetica")
        sns.set_style("ticks")
        if tech == 'PCA':
            retech = PCA(n_components=2)
            embeddings = retech.fit_transform(np.array(X))

        elif tech == 'TSNE':
            retech = TSNE(n_components=2)
            embeddings = retech.fit_transform(np.array(X))

        elif tech == 'UMAP':
            import umap
            retech = umap.UMAP(n_components=2)
            embeddings = retech.fit_transform(np.array(X))

        else:
            embeddings = np.array(X)
        fig, ax = plt.subplots(1, 1, figsize=(10, 8))
        sp = ax.scatter(
            embeddings[:, 0], embeddings[:, 1],
            c=y,
            s=marker_size,
            cmap=plt.cm.get_cmap(cmap, 10),
        )
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.set_xlabel(tech + ' 1', fontsize=14)
        ax.set_ylabel(tech + ' 2', fontsize=14)
        ax.set_title(title, fontsize=14)
        fig.colorbar(sp)
        fig.subplots_adjust(
            top=0.95,
            bottom=0.1,
            left=0.1,
            right=0.98,
            # hspace=0.20,
            # wspace=0.15
        )
        plt.show()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from typing import Dict

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from umiche.deduplicate.TMR import TMR


class TripletErrorCode:

    def __init__(self, error_rate):
        self.error_rate = error_rate
        self.tmr = TMR(error_rate=self.error_rate)

        sns.set(font="Helvetica")
        sns.set_style("ticks")

    def incorrect(
            self,
            num_nt=12,
    ):
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(7, 5), sharey=True, sharex=False)
        self.tmr.set_num_nt(num_nt)
        P_std_umi_error = self.tmr.monomer_umi_error
        P_homo_umi_error = self.tmr.homotrimer_umi_error
        P_homo_umi_error_upper_bound = self.tmr.bitflip_code_error
        ax.grid(False)
        ax.plot(
            self.error_rate,
            P_std_umi_error,
            label='Standard 12-bp UMI',
            color="slategrey",
            linewidth=2.5,
        )
        ax.plot(
            self.error_rate,
            P_homo_umi_error,
            label='Homotrimer 36-bp UMI',
            color="crimson",
            linewidth=2.5,
        )
        ax.plot(
            self.error_rate,
            P_homo_umi_error_upper_bound,
            label='36 binary repetition codes (n=3)',
            # label='36-size binary codes with the 3-input majority gate',
            color="olivedrab",
            linewidth=2.5,
        )
        ax.spines['right'].set_color('none')
        ax.spines['top'].set_color('none')
        ax.tick_params(axis='both', which='major', labelsize=16)
        ax.tick_params(axis='both', which='minor', labelsize=16)
        ax.set_xlabel("Single-base error rate ($p$)", fontsize=16)
        ax.set_ylabel("Probability of incorrect synthesis", fontsize=16)
        # ax.set_title("Comparison of UMI error probability vs. per-base error rate", fontsize=14)
        ax.legend(fontsize=14, frameon=False)
        plt.tight_layout()
        plt.show()
        return

    def correct(
            self,
            num_nt=1,
    ):
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(7, 5), sharey=True, sharex=False)
        self.tmr.set_num_nt(num_nt)
        P_std_umi_correct = self.tmr.monomer_umi_error_free
        P_homo_umi_correct = self.tmr.homotrimer_umi_error_free
        P_homo_umi_correct_upper_bound = self.tmr.bitflip_code_error_free
        ax.grid(False)
        ax.plot(
            self.error_rate,
            P_std_umi_correct,
            label='A single base',
            color="slategrey",
            linewidth=2.5,
        )
        ax.plot(
            self.error_rate,
            P_homo_umi_correct,
            label='Homotrimer block',
            color="crimson",
            linewidth=2.5,
        )
        ax.plot(
            self.error_rate,
            P_homo_umi_correct_upper_bound,
            label='Binary repetition code block (n=3)',
            # label='3-input majority gate',
            color="olivedrab",
            linewidth=2.5,
        )
        ax.spines['right'].set_color('none')
        ax.spines['top'].set_color('none')
        ax.tick_params(axis='both', which='major', labelsize=16)
        ax.tick_params(axis='both', which='minor', labelsize=16)
        ax.set_xlabel("Single-base error rate ($p$)", fontsize=16)
        ax.set_ylabel("Probability of correct synthesis", fontsize=16)
        # ax.set_title("building block", fontsize=14)
        ax.legend(fontsize=14, frameon=False)
        plt.tight_layout()
        plt.show()
        return


if __name__ == "__main__":
    e = np.linspace(0.00001, 0.5, 500)
    p = TripletErrorCode(error_rate=e)
    # p.incorrect(num_nt=12)
    p.correct(num_nt=1)


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


import argparse
from umiche.trim.Template import Template as umitrim
from umiche.util.Console import Console


class fixed:

    def __init__(self, mode='external', params=None, verbose=True):
        if mode == 'internal':
            self.params = params
            self.verbose = verbose
            self.console = Console()
            self.console.verbose = self.verbose
            self.console.print('run Mclumi internally.')
            self.console.print('Your params for trimming UMIs are: \n{}'.format(self.params))
        else:
            self.parser = argparse.ArgumentParser(description='Sequence Identity calculations')
            self.parser.add_argument(
                "--read_structure", "-rs",
                metavar='read_structure',
                dest='rs',
                required=True,
                type=str,
                help='the read structure with elements in conjunction with +',
            )
            self.parser.add_argument(
                "--lens", "-l",
                metavar='lens',
                dest='l',
                required=True,
                type=str,
                help='lengths of all sub-structures separated by +',
            )
            self.parser.add_argument(
                "--input", "-i",
                metavar='input',
                dest='i',
                required=True,
                type=str,
                help='input a fastq file in gz format for trimming UMIs',
            )
            self.parser.add_argument(
                "--output", "-o",
                metavar='output',
                dest='o',
                required=True,
                type=str,
                help='output a UMI-trimmed fastq file in gz format.',
            )
            self.parser.add_argument(
                "--verbose", "-vb",
                metavar='verbose',
                dest='vb',
                default=True,
                type=bool,
                help='to enable if output logs are on console',
            )
            args = self.parser.parse_args()
            self.verbose = args.vb
            self.read_structure = args.rs
            self.structure_lengths = []
            for i in args.l.split('+'):
                self.structure_lengths.append(int(i))
            # print(args.l)
            # print(args.i)
            self.fastq_fpn = args.i
            self.fastq_trimmed_fpn = args.o
            self.params = {s: {'len': l} for s, l in zip(self.read_structure.split('+'), self.structure_lengths)}
            self.params['read_struct'] = self.read_structure
            self.params['fastq'] = {
                'fpn': self.fastq_fpn,
                'trimmed_fpn': self.fastq_trimmed_fpn,
            }
            self.console = Console()
            self.console.verbose = self.verbose

    def call(self, ):
        umitrim_parser = umitrim(self.params, self.verbose)
        df = umitrim_parser.todf()
        umitrim_parser.togz(df)


if __name__ == "__main__":
    from umiche.path import to

    params = {
        'umi_1': {
            'len': 12,
        },
        'umi_2': {
            'len': 10,
        },
        'umi_3': {
            'len': 12,
        },
        # 'read_struct': 'umi_1+seq_1',
        'read_struct': 'primer_1+umi_1+seq_1+seq_2+umi_2+umi_3+primer_2',
        # 'read_struct': 'primer+umi_start+seq+umi_1+primer',
        # 'read_struct': 'primer+umi_1+seq_1+primer+umi_2+seq_2+umi_3+primer',
        'primer_1': {
            'len': 20,
        },
        'primer_2': {
            'len': 20,
        },
        'seq_1': {
            'len': 6,
        },
        'seq_2': {
            'len': 8,
        },
        'fastq': {
            'fpn': to('example/data/pcr_1.fastq.gz'),
            'trimmed_fpn': to('example/data/pcr_1_trim.fastq.gz'),
        },
    }
    p = fixed(
        mode='internal',
        # mode='external',

        params=params,
    )
    p.call()


__version__ = "v1.0"
__copyright__ = "Copyright 2025"
__license__ = "GPL-3.0"
__developer__ = "Jianfeng Sun"
__maintainer__ = "Jianfeng Sun"
__email__="jianfeng.sunmt@gmail.com"


from umiche.deduplicate.method.trimer.Collapse import Collapse
from umiche.util.Console import Console


class MajorityVote:

    def __init__(
            self,
            verbose=False,
    ):
        self.collapse = Collapse()

        self.console = Console()
        self.console.verbose = verbose

    def track(
            self,
            multimer_list,
            recur_len,
    ):
        """

        Parameters
        ----------
        multimer_list

        Returns
        -------

        """
        multimer_umi_to_mono_umi_map = {multimer_umi: self.collapse.majority_vote(
            umi=multimer_umi,
            recur_len=recur_len,
        ) for multimer_umi in multimer_list}
        mono_umi_to_multimer_umi_map = {self.collapse.majority_vote(
            umi=multimer_umi,
            recur_len=recur_len,
        ): multimer_umi for multimer_umi in multimer_list}

        uniq_multimer_cnt = len(multimer_umi_to_mono_umi_map)

        shortlisted_multimer_umi_list = [*mono_umi_to_multimer_umi_map.values()]
        dedup_cnt = len(shortlisted_multimer_umi_list)
        self.console.print('=========># of shortlisted multimer UMIs: {}'.format(len(shortlisted_multimer_umi_list)))
        self.console.print('=========>dedup cnt: {}'.format(dedup_cnt))
        return dedup_cnt, uniq_multimer_cnt, shortlisted_multimer_umi_list


if __name__ == "__main__":
    from umiche.path import to

    p = MajorityVote()